#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Functions for parsing a config file for dbgen environmental variables"""
import pathlib

# External Imports
import shlex
import subprocess
import sys
import tempfile
from collections import OrderedDict
from configparser import ConfigParser
from dataclasses import dataclass, field, fields
from logging import getLogger
from os import environ
from os.path import dirname, expanduser, expandvars, join
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

# Internal imports
from dbgen.utils.exceptions import DBgenConfigException
from dbgen.utils.module_loading import import_string

logger = getLogger("dbgen.config")


def _get_config_value_from_secret_backend(config_key):
    """Get Config option values from Secret Backend"""
    # !TODO!
    raise NotImplementedError
    secrets_client = None  # get_custom_secret_backend()
    if not secrets_client:
        return None
    return secrets_client.get_config(config_key)


def _read_default_config_file(file_name: str) -> Tuple[str, str]:
    templates_dir = join(dirname(__file__), "config_templates")
    file_path = join(templates_dir, file_name)
    with open(file_path, encoding="utf-8") as config_file:
        return config_file.read(), file_path


DEFAULT_CONFIG, DEFAULT_CONFIG_FILE_PATH = _read_default_config_file("default_dbgen.cfg")


def expand_env_var(env_var):
    """
    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.
    """
    if not env_var:
        return env_var
    while True:
        interpolated = expanduser(expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


def run_command(command):
    """Runs command and returns stdout"""
    process = subprocess.Popen(
        shlex.split(command),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
    )
    output, stderr = [stream.decode(sys.getdefaultencoding(), "ignore") for stream in process.communicate()]

    if process.returncode != 0:
        raise DBgenConfigException(
            f"Cannot execute {command}. Error code is: {process.returncode}. "
            f"Output: {output}, Stderr: {stderr}"
        )

    return output


class DBgenConfigParser(ConfigParser):
    """
    Parser for the dbgen section in a .ini formatted config file.

    Args:
        configparser ([type]): [description]
    """

    sensitive_config_values = {
        ("core", "db_conn_str"),
        ("core", "metadb_conn_str"),
        ("logging", "log_path"),
    }

    deprecated_options: Dict[Tuple[str, str], Tuple[str, str, str]] = {}
    ENV_VAR_PREFIX = "DBGEN__"

    def __init__(self, default_config=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.dbgen_defaults = ConfigParser(*args, **kwargs)
        if default_config is not None:
            self.dbgen_defaults.read_string(default_config)

        self.is_validated = False

    def _env_var_name(self, section: str, key: str) -> str:
        return f"{self.ENV_VAR_PREFIX}{section.upper()}__{key.upper()}"

    def _get_env_var_option(self, section, key):
        # must have format DBGEN__{SECTION}__{KEY} (note double underscore)
        env_var = self._env_var_name(section, key)
        if env_var in environ:
            return expand_env_var(environ[env_var])
        # alternatively DBGEN__{SECTION}__{KEY}_CMD (for a command)
        env_var_cmd = env_var + "_CMD"
        if env_var_cmd in environ:
            # if this is a valid command key...
            if (section, key) in self.sensitive_config_values:
                return run_command(environ[env_var_cmd])
        # alternatively DBGEN__{SECTION}__{KEY}_SECRET (to get from Secrets Backend)
        env_var_secret_path = env_var + "_SECRET"
        if env_var_secret_path in environ:
            # if this is a valid secret path...
            if (section, key) in self.sensitive_config_values:
                return _get_config_value_from_secret_backend(environ[env_var_secret_path])
        return None

    def _get_cmd_option(self, section, key):
        fallback_key = key + "_cmd"
        # if this is a valid command key...
        if (section, key) in self.sensitive_config_values:
            if super().has_option(section, fallback_key):
                command = super().get(section, fallback_key)
                return run_command(command)
        return None

    def _get_secret_option(self, section, key):
        """Get Config option values from Secret Backend"""
        fallback_key = key + "_secret"
        # if this is a valid secret key...
        if (section, key) in self.sensitive_config_values:
            if super().has_option(section, fallback_key):
                secrets_path = super().get(section, fallback_key)
                return _get_config_value_from_secret_backend(secrets_path)
        return None

    def get(self, section, key, **kwargs):
        section = str(section).lower()
        key = str(key).lower()

        deprecated_section, deprecated_key, _ = self.deprecated_options.get(
            (section, key), (None, None, None)
        )

        option = self._get_environment_variables(deprecated_key, deprecated_section, key, section)
        if option is not None:
            return option

        option = self._get_option_from_config_file(deprecated_key, deprecated_section, key, kwargs, section)
        if option is not None:
            return option

        option = self._get_option_from_commands(deprecated_key, deprecated_section, key, section)
        if option is not None:
            return option

        option = self._get_option_from_secrets(deprecated_key, deprecated_section, key, section)
        if option is not None:
            return option

        return self._get_option_from_default_config(section, key, **kwargs)

    def _get_option_from_default_config(self, section, key, **kwargs):
        # ...then the default config
        if self.dbgen_defaults.has_option(section, key) or "fallback" in kwargs:
            return expand_env_var(self.dbgen_defaults.get(section, key, **kwargs))

        else:
            logger.warning("section/key [%s/%s] not found in config", section, key)

            raise DBgenConfigException(f"section/key [{section}/{key}] not found in config")

    def _get_option_from_secrets(self, deprecated_key, deprecated_section, key, section):
        # ...then from secret backends
        option = self._get_secret_option(section, key)
        if option:
            return option
        if deprecated_section:
            option = self._get_secret_option(deprecated_section, deprecated_key)
            if option:
                self._warn_deprecate(section, key, deprecated_section, deprecated_key)
                return option
        return None

    def _get_option_from_commands(self, deprecated_key, deprecated_section, key, section):
        # ...then commands
        option = self._get_cmd_option(section, key)
        if option:
            return option
        if deprecated_section:
            option = self._get_cmd_option(deprecated_section, deprecated_key)
            if option:
                self._warn_deprecate(section, key, deprecated_section, deprecated_key)
                return option
        return None

    def _get_option_from_config_file(self, deprecated_key, deprecated_section, key, kwargs, section):
        # ...then the config file
        if super().has_option(section, key):
            # Use the parent's methods to get the actual config here to be able to
            # separate the config from default config.
            return expand_env_var(super().get(section, key, **kwargs))
        if deprecated_section:
            if super().has_option(deprecated_section, deprecated_key):
                self._warn_deprecate(section, key, deprecated_section, deprecated_key)
                return expand_env_var(super().get(deprecated_section, deprecated_key, **kwargs))
        return None

    def _get_environment_variables(self, deprecated_key, deprecated_section, key, section):
        # first check environment variables
        option = self._get_env_var_option(section, key)
        if option is not None:
            return option
        if deprecated_section:
            option = self._get_env_var_option(deprecated_section, deprecated_key)
            if option is not None:
                self._warn_deprecate(section, key, deprecated_section, deprecated_key)
                return option
        return None

    def getsection(self, section: str) -> Optional[Dict[str, Union[str, int, float, bool]]]:
        """
        Returns the section as a dict. Values are converted to int, float, bool
        as required.

        :param section: section from the config
        :rtype: dict
        """
        if not self.has_section(section) and not self.dbgen_defaults.has_section(section):
            return None

        if self.dbgen_defaults.has_section(section):
            _section = OrderedDict(self.dbgen_defaults.items(section))
        else:
            _section = OrderedDict()

        if self.has_section(section):
            _section.update(OrderedDict(self.items(section)))

        section_prefix = self._env_var_name(section, "")
        for env_var in sorted(environ.keys()):
            if env_var.startswith(section_prefix):
                key = env_var.replace(section_prefix, "")
                if key.endswith("_CMD"):
                    key = key[:-4]
                key = key.lower()
                _section[key] = self._get_env_var_option(section, key)

        for key, val in _section.items():
            try:
                val = int(val)
            except ValueError:
                try:
                    val = float(val)
                except ValueError:
                    if val.lower() in ("t", "true"):
                        val = True
                    elif val.lower() in ("f", "false"):
                        val = False
            _section[key] = val
        return _section

    def getimport(self, section, key, **kwargs):  # noqa
        """
        Reads options, imports the full qualified name, and returns the object.

        In case of failure, it throws an exception a clear message with the key aad the section names

        :return: The object or None, if the option is empty
        """
        full_qualified_path = self.get(section=section, key=key, **kwargs)
        if not full_qualified_path:
            return None

        try:
            return import_string(full_qualified_path)
        except ImportError as e:
            logger.error(e)
            raise DBgenConfigException(
                f'The object could not be loaded. Please check "{key}" key in "{section}" section. '
                f'Current value: "{full_qualified_path}".'
            )


@dataclass
class RunConfig:
    """Captures all configurations for a run"""

    only: List[str] = field(default_factory=lambda: [])
    xclude: List[str] = field(default_factory=lambda: [])
    add: bool = False
    retry: bool = False
    start: Optional[str] = None
    until: Optional[str] = None
    serial: bool = False
    nuke: bool = False
    bar: bool = True
    skip_row_count: bool = True
    batch: Optional[int] = None

    @property
    def fields(self):
        return set(map(lambda x: x.name, fields(self)))


DEFAULT_CONFIG, DEFAULT_CONFIG_FILE_PATH = _read_default_config_file("default_dbgen.cfg")


def get_dbgen_home():
    """Get path to DBGEN Home"""
    return expand_env_var(environ.get("DBGEN_HOME", "~/.dbgen"))


def get_dbgen_config(dbgen_home: str):
    """Get Path to airflow.cfg path"""
    if "DBGEN_CONFIG" not in environ:
        return join(dbgen_home, "dbgen.cfg")
    return expand_env_var(environ["DBGEN_CONFIG"])


DBGEN_HOME = get_dbgen_home()
DBGEN_CONFIG = get_dbgen_config(DBGEN_HOME)
pathlib.Path(DBGEN_HOME).mkdir(parents=True, exist_ok=True)


def parameterized_config(template):
    """
    Generates a configuration from the provided template + variables defined in
    current scope

    :param template: a config content templated with {{variables}}
    """
    all_vars = {k: v for d in [globals(), locals()] for k, v in d.items()}
    return template.format(**all_vars)  # noqa


def get_config(config_file: Path) -> DBgenConfigParser:
    config = DBgenConfigParser(default_config=parameterized_config(DEFAULT_CONFIG))
    if not config_file.exists():
        TEMPLATE_START = '# ----------------------- TEMPLATE BEGINS HERE -----------------------'
        logger.info('Creating new DBgen config file in: %s', config_file)
        with open(config_file, 'w') as file:
            cfg = parameterized_config(DEFAULT_CONFIG)
            cfg = cfg.split(TEMPLATE_START)[-1].strip()
            file.write(cfg)
    # Read the path if it exists
    if config_file:
        config.read(config_file.absolute())
    else:
        logger.debug(f"No Config file provided. Using default values for DBgen variables")
    return config


config = get_config(Path(DBGEN_CONFIG))
# TODO DEFAULT_ENV and DBGEN_TMP cannot be overwritten at runtime must be stored in env var or config file
# Would be nice to make these changeable at runtime
DBGEN_TMP_STR = environ.get("DBGEN_TMP") or config.get("core", "dbgen_tmp", fallback=tempfile.gettempdir())
DBGEN_TMP = Path(DBGEN_TMP_STR)
DBGEN_TMP.mkdir(exist_ok=True, parents=True)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from base64 import b64encode
from hashlib import md5

# External
from typing import TYPE_CHECKING
from typing import Dict as D
from typing import List as L
from typing import Sequence
from typing import Set as S
from typing import Tuple as T
from typing import Union as U

from dbgen.utils.lists import flatten
from dbgen.utils.misc import Base

# Internal
if TYPE_CHECKING:
    from dbgen.core.schema import Entity, Rel, RelTup, SuperRel

    Rel, SuperRel, RelTup, Entity
################################################################################


class Path(Base):
    """
    (loop 2x)
                 ->           ^|
            -> C -> D         |v
    A -> B            -> E -> F -> G
              -> H

    Would be represented like so:
    Path('G',[fg,ff,ff,ef,[[de,[[cd1,bc,ab],
                                [cd2,bc,ab]],
                           [he,bh,ab]
                          ]])
    """

    def __init__(self, end: U[str, "Entity"], fks: list = None, name: str = None) -> None:
        self.end = end if isinstance(end, str) else end.name
        self.fks = fks or []
        self.name = name
        err = "expected {} in {} (objs of {})\nall fks: {}"

        if fks and fks[0]:
            if isinstance(fks[0], list):
                for fk in fks[0]:
                    assert self.end in fk[0].objs, err.format(self.end, fk[0].objs, fk[0], self.fks)
            else:
                assert self.end in fks[0].objs, err.format(self.end, fks[0].objs, fks[0], self.fks)
        super().__init__()

    def __str__(self) -> str:
        return str(self.join())

    def __repr__(self) -> str:
        return f'JPath("{self.end}", {self.fks})'

    def __add__(self, other: "Path") -> "Path":
        """
        Concatenate paths: tail of first must match base of second

        P1 (A --> B), P2 (B --> C) ==> P1 + P2 (A --> C)
        """
        assert other.linear, "Cannot concatenate paths if the second path branches"
        assert self.end == other.base, "Cannot concatenate paths unless head/tail matches"
        assert self.name == other.name, "Cannot concatenate paths with different names"
        return Path(other.end, other.fks + self.fks, name=self.name)

    def __sub__(self, other: "Path") -> "Path":
        """
        Take a path difference to truncate edges from the start of a path
            ... only defined under rare circumstances

        P1 (A --> C), P2 (A --> B) ==> P1 - P2 (B --> C)
        """
        l2 = len(other.fks)
        err = "Cannot take path difference: latter path is not subset of first"
        assert self.fks[-l2:] == other.fks, err
        assert self.name == other.name, "Cannot take path difference with different names"
        return Path(self.end, self.fks[:l2], name=self.name)

    @property
    def linear(self) -> bool:
        return all([not isinstance(fk, list) for fk in self.fks])

    @property
    def base(self) -> str:
        """The start of the current join path, defined iff it is linear"""
        assert self.linear
        curr = self.end
        for fk in self.fks:
            curr = fk.other(curr)
        return curr

    def all_rels(self) -> S["Rel"]:
        stack = self.fks
        out: S["SuperRel"] = set()
        self.newmethod257(stack, out)
        return {o.to_rel() for o in out}

    def newmethod257(self, stack, out):
        while stack:
            curr = stack.pop()

            if isinstance(curr, list):
                assert not stack
                stack = flatten(curr)
            else:
                out.add(curr)

    def add(self, r: "Rel") -> "Path":
        assert self.base in r.objs
        return Path(self.end, self.fks + [r])

    def add_branch(self, p: "Path") -> "Path":
        c = self.copy()  # only work with a deepcopy

        # Check if this is the first branch we're adding
        if not c.fks or not isinstance(c.fks[-1], list):
            c.fks.append([p.fks])
        else:
            c.fks[-1].append(p.fks)
        return c

    def join(self) -> "Join":
        """Get top-level join that is implied by this path."""
        j = Join(self.end, name=self.name)
        if self.fks:
            nextfk = self.fks[0]
            if isinstance(nextfk, list):
                for nex in nextfk:
                    p = Path(nex[0].other(self.end), nex[1:], name=self.name)
                    j.add(p.join(), nex[0])
            else:
                nextab = nextfk.other(self.end)
                nextpath = Path(nextab, self.fks[1:], name=self.name)
                j.add(nextpath.join(), nextfk)
        return j

    def alljoin(self) -> S["Join"]:
        stack = [self.join()]
        joins = set()  # type: S['Join']
        while stack:
            curr = stack.pop()
            if curr not in joins:
                joins.add(curr)
                stack.extend(list(curr.conddict.keys()))
        return joins

    def _from(self) -> "From":
        return From(joins=self.alljoin())


class Join(Base):
    """
    Constructed from a Path object

    Represents a table plus a unique set of JOINs that led to that table
        (accounting for multiple linear paths)
    """

    def __init__(self, obj: str, conds: L[T["Join", S["SuperRel"]]] = None, name: str = None) -> None:
        assert isinstance(obj, str)
        self.obj = obj
        self.conds = conds or []
        self.name = name
        super().__init__()

    def __str__(self) -> str:
        return self.alias

    def __repr__(self) -> str:
        return f"JOIN {self.obj} ({','.join(map(str, self.conds))})"

    def __lt__(self, other: "Join") -> bool:
        return str(self) < str(other)

    @property
    def conddict(self) -> D["Join", S["SuperRel"]]:
        return dict(self.conds)

    # Public Methods
    def add(self, j: "Join", e: "SuperRel") -> None:
        if j in self.conddict:
            d = dict(self.conddict)
            d[j].add(e)
            self.conds = list(d.items())  # overwrite
        else:
            self.conds.append((j, {e}))

    @property
    def alias(self) -> str:
        """How to uniquely refer to this PATH through the schema"""

        if not self.conds:
            return self.obj

        s = ""
        for j, fks in sorted(self.conds):
            fkstr = "|".join([fk.print() for fk in sorted(fks)])
            s += f"[({fkstr})#({j.alias})]"

        data = s + self.obj
        m = md5(data.encode("ascii"))
        out = b64encode(m.digest()).decode("ascii")[:3]
        out_name = f"{self.name}-" if self.name is not None else ""
        return self.obj + f"({out_name}{out})"

    def print(self, optional: L["RelTup"] = None) -> str:
        """Render JOIN statement in FROM clause"""
        conds = [self._cond(j, e) for j, e in self.conds]  # conditions to join on
        opts = optional or []
        if not bool(conds):
            jointype = " CROSS "
        else:
            left = True
            # Assume a left join. if any FKs in current edge are NOT in
            # "optional", then set to Inner join
            for e in self.conddict.values():
                for fk in e:  # type: ignore
                    if fk.tup() not in opts:
                        left = False
                        break
            jointype = " LEFT " if left else " INNER "
        on = (
            " ON \n\t\t\t" + "\n\t\t\tAND ".join(conds) if conds else ""
        )  # Possibly do not join on any condition
        args = [jointype, self.obj, self.alias, on]
        return '\n\t{}JOIN {} AS "{}" {}'.format(*args)

    # Private Methods

    def _cond(self, j: "Join", rels: S["SuperRel"]) -> str:
        """Assume the alias defined by the arg's Join has already been defined
        in the FROM statement. Write a SQL JOIN condition that will be used
        to define the current Join object's alias

        NEED TO MODIFY EVERYTHING SO THAT WE USE <object>.id, not <object>.name + '_id'
        """

        conds = []  # type: L[str]
        for fk in rels:
            o = fk.other(self.obj)
            forward = o == fk.source  # Rel in forward direction. Self.obj is the 'old table'
            aliases = [j.alias, self.alias]
            cols = [fk.name, fk.target_id_str] if forward else [fk.target_id_str, fk.name]
            args = [aliases[0], cols[0], aliases[1], cols[1]]
            new = ' "{}"."{}" = "{}"."{}" '.format(*args)
            conds.append(new)
        return "\n\t\tAND ".join(conds)


class From(Base):
    """
    Class used to help construct a WHERE clause.
    """

    def __init__(self, basis: L[str] = None, joins: S[Join] = None) -> None:
        self.joins = {Join(b) for b in basis or []} | (joins or set())
        self.basis = {j.obj for j in self.joins if not j.conds}
        super().__init__()

    # @property --- IS THIS GOING TO BREAK STUFF NOW THAT IT'S A FIELD
    # def basis(self) -> S[str]: return {j.obj for j in self.joins if not j.conds}

    # Public methods #
    def __str__(self) -> str:
        return "From(basis=%s,%d joins)" % (self.basis, len(self.joins))

    def __or__(self, f: "From") -> "From":
        return From(joins=self.joins | f.joins)

    def print(self, optional: Sequence["RelTup"] = None) -> str:
        from dbgen.utils.graphs import DiGraph, topsort_with_dict

        d = {j.alias: j for j in self.joins}
        G = DiGraph()
        G.add_nodes_from(d.keys())
        for j in self.joins:
            for j2 in j.conddict.keys():
                G.add_edge(j.alias, j2.alias)
        sort = list(reversed(topsort_with_dict(G, d)))
        start = sort[0].obj
        assert start in self.basis
        return "FROM " + start + "".join([j.print(optional) for j in sort[1:]])

    def aliases(self) -> L[str]:
        return [j.alias for j in self.joins]

    def pks(self, agg: bool = False) -> str:
        col = 'MAX("{0}"."{1}_id")' if agg else '"{0}"."{1}_id"'
        return ",\n\t".join(
            [(col + ' AS "{0}" ').format(a, j.obj) for a, j in sorted(zip(self.aliases(), self.joins))]
        )



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Define the Load Object for inserting transformed data into the Database."""
import logging
import re
from io import StringIO
from random import getrandbits

# External Modules
from typing import TYPE_CHECKING
from typing import Dict as D
from typing import List as L
from typing import Tuple as T
from typing import Union as U

import psycopg2
from hypothesis.strategies import SearchStrategy, builds, dictionaries, just, one_of
from jinja2 import Template
from psycopg2.errors import QueryCanceled

from dbgen.core.funclike import Arg, ArgLike, Const
from dbgen.utils.exceptions import DBgenExternalError, DBgenTypeError, Psycopg2Error
from dbgen.utils.lists import broadcast
from dbgen.utils.misc import Base, nonempty
from dbgen.utils.sql import Connection as Conn
from dbgen.utils.str_utils import hashdata_

# Internal Modules
if TYPE_CHECKING:
    from dbgen.core.model.model import UNIVERSE_TYPE
    from dbgen.core.schema import Entity

"""
Defines the class of modifications to a database

There is a horrific amount of duplicated code in this file...... oughta fixit
"""
################################################################################
# ######################
# # Constants
# # --------------------
NUM_QUERY_TRIES = 3


class Load(Base):
    """
    The purpose for this object is to make an easily serializable data structure
    that knows how to update the database (these methods could easily be for
    Model, but we don't want to send the entire model just to do this small thing)
    """

    def __init__(
        self,
        obj: str,
        attrs: D[str, ArgLike],
        fks: D[str, "Load"],
        pk: U[Arg, Const] = None,
        insert: bool = False,
        partition_attr: str = None,
    ) -> None:
        """
        Initializes Load object with relevant Objects and nested Loads. This is
        not intended to be called by users, but rather called through the
        Entity.__call__ method (Entity()).

        Args:
            obj (str): [description]
            attrs (D[str, ArgLike]): [description]
            fks (D[str,): [description]
            pk (Arg, optional): [description]. Defaults to None.
            insert (bool, optional): [description]. Defaults to False.
        """
        self.obj = obj.lower()
        self.attrs = {k.lower(): v for k, v in attrs.items()}
        self.fks = {k.lower(): v for k, v in fks.items()}
        self.pk = pk
        self.insert = insert
        self.partition_attr = partition_attr
        self._logger = logging.getLogger(f"dbgen.run.loading.{self.obj}")
        self._logger.setLevel(logging.DEBUG)
        err = "Cant insert %s if we already have PK %s"
        assert (pk is None) or (not insert), err % (obj, pk)
        if isinstance(pk, Const) and pk == Const(None):
            pass
        else:
            assert isinstance(pk, (Arg, type(None))), (
                obj,
                attrs,
                fks,
                pk,
                insert,
            )
        super().__init__()

    def __str__(self) -> str:
        n = len(self.attrs)
        m = len(self.fks)
        return "Load<%s, %d attr, %d rel>" % (self.obj, n, m)

    ##################
    # Public methods #
    ###################
    def tabdeps(self, universe: D[str, "Entity"]) -> L[str]:
        """All tables that are updated (they must already exist, is the logic)"""
        from dbgen.core.schema import Partition

        # Check if we are updating; if so we depend on
        entity = universe[self.obj]
        out = [entity.name] if not self.insert else []

        # Check for partitions on self table
        # If so add partitions if Entity and add parent entity if partition
        if out:
            if isinstance(entity, Partition):
                out.append(entity._parent_name)
            elif entity.partition_attr:
                partition_names = map(lambda x: x.name, entity.get_all_partitions())
                out.extend(partition_names)

        for fk in self.fks.values():
            # Check if we are simply setting the FK to null
            if not isinstance(fk.pk, Const) or fk.pk != Const(None):
                out.extend(fk.tabdeps(universe))
        return out

    def newtabs(self, universe: D[str, "Entity"]) -> L[str]:
        """All tables that could be inserted into this load"""
        from dbgen.core.schema import Partition

        entity = universe[self.obj]
        out = [entity.name] if self.insert else []

        # Check for partitions on self table
        # If so add partitions if Entity and add parent entity if partition
        if out and isinstance(entity, Partition):
            out.append(entity._parent_name)
        elif out and entity.partition_attr:
            partition_names = map(lambda x: x.name, entity.get_all_partitions())
            out.extend(partition_names)

        for a in self.fks.values():
            out.extend(a.newtabs(universe))
        return out

    def newcols(self, universe: D[str, "Entity"]) -> L[str]:
        """All attributes that could be populated by this load"""
        from dbgen.core.schema import Partition

        entity = universe[self.obj]
        if isinstance(entity, Partition):
            new_entities = [entity, universe[entity._parent_name]]
        else:
            new_entities = [entity, *entity.get_all_partitions()]
        out: L[str] = []
        for new_entity in new_entities:
            out.extend(
                [
                    new_entity.name + "." + a
                    for a in self.attrs.keys()
                    if (self.insert or (a not in new_entity.ids()))
                ]
            )
            for k, a in self.fks.items():
                if self.insert or (k not in new_entity.id_fks()):
                    out.extend([new_entity.name + "." + k] + a.newcols(universe))
        return out

    def act(
        self,
        cxn: Conn,
        universe: "UNIVERSE_TYPE",
        rows: L[dict],
        gen_name: str,
    ) -> None:
        """
        Top level call from a Generator to execute an load (top level is
        always insert or update, never just a select)
        """
        # Initialize logger
        self._load(cxn, universe, rows, insert=self.insert)

    def rename_object(self, o: "Entity", n: str) -> "Load":
        """Replaces all references to a given object to one having a new name"""
        a = self.copy()
        if a.obj == o.name:
            a.obj = n
        for k, v in a.fks.items():
            a.fks[k] = v.rename_object(o, n)
        return a

    @classmethod
    def _strat(cls) -> SearchStrategy:
        """A hypothesis strategy for generating random examples."""

        common_load_kwargs = dict(
            obj=nonempty,
            attrs=dictionaries(keys=nonempty, values=ArgLike._strat()),
        )
        load_ = builds(
            cls,
            fks=just(dict()),
            pk=Arg._strat(),
            insert=just(False),
            **common_load_kwargs,  # type: ignore
        )
        load0 = builds(
            cls,
            fks=dictionaries(keys=nonempty, values=load_),
            pk=Arg._strat(),
            insert=just(False),
            **common_load_kwargs,  # type: ignore
        )
        load1 = builds(
            cls,
            fks=dictionaries(keys=nonempty, values=load_),
            pk=just(None),
            insert=just(True),
            **common_load_kwargs,  # type: ignore
        )
        return one_of(load_, load0, load1)

    ###################
    # Private methods #
    ###################

    def _getvals(
        self,
        universe: "UNIVERSE_TYPE",
        row: dict,
    ) -> T[L[int], L[tuple]]:
        """
        Get a broadcasted list of INSERT/UPDATE values for an object, given
        Pyblock+Query output
        """

        idattr, allattr = [], []
        obj_pk_name, ids, id_fks, dtype_dict = universe[self.obj]
        for k, v in sorted(
            self.attrs.items(),
        ):
            val = v.arg_get(row)
            dtype = dtype_dict[k]
            if isinstance(val, (list, tuple)):
                func = lambda x: [dtype.cast(ele) for ele in x]
            else:
                func = dtype.cast
            try:
                allattr.append(func(val))
            except DBgenTypeError:
                self._logger.error(
                    f"While processing a row we encountered a value that violated column {k}'s type constraint {dtype}."
                    f"\nRow: {row}\nVal: {val}"
                )
                raise
            if k in ids:
                idattr.append(allattr[-1])

        for kk, vv in sorted(self.fks.items()):
            if vv.pk is not None:
                val = vv.pk.arg_get(row)
                if isinstance(val, int) or val is None:
                    pass
                elif isinstance(val, list) and isinstance(val[0], int):
                    pass
                else:
                    raise ValueError(f"Primary Key is not an int or None: {row}\n{vv}\n{vv.pk}\n{val}")
            else:
                val, fk_adata = vv._getvals(universe, row)

            allattr.append(val)
            if kk in id_fks:
                idattr.append(val)

        idata: L[tuple] = broadcast(idattr)
        adata: L[tuple] = broadcast(allattr)
        if self.pk is not None:
            assert not idata, "Cannot provide a PK *and* identifying info"
            pkdata = self.pk.arg_get(row)
            if isinstance(pkdata, int):
                idata_prime = [pkdata]
            elif isinstance(pkdata, list) and isinstance(pkdata[0], int):  # HACKY
                idata_prime = pkdata
            elif isinstance(pkdata, str) or (isinstance(pkdata, list) and isinstance(pkdata[0], str)):
                raise TypeError(
                    f"While looking for the PK on {self.obj}, we found a string or list of strings: {pkdata}\n"
                    "PK's should be integers for hashing purposes."
                )
            else:
                raise TypeError(
                    "PK should either receive an int or a list of ints",
                    vars(self),
                )
        else:
            idata_prime = []
            idata_dict = {}  # type: D[tuple,int]
            # Iterate through the identifying data and cache the hashed result for speed
            for idata_curr in idata:
                if idata_dict.get(idata_curr) is None:
                    try:
                        idata_dict[idata_curr] = int(hashdata_(idata_curr))
                    except TypeError:
                        print(idata_curr)
                idata_prime.append(idata_dict[idata_curr])

        if len(idata_prime) == 1:
            idata_prime *= len(adata)  # broadcast

        lenerr = "Cannot match IDs to data: %d!=%d"
        assert len(idata_prime) == len(adata), lenerr % (
            len(idata_prime),
            len(adata),
        )
        return idata_prime, adata

    def _data_to_stringIO(
        self,
        pk: L[int],
        data: L[tuple],
        obj_pk_name: str,
    ) -> StringIO:
        """
        Function takes in a path to a delimited file and returns a IO object
        where the identifying columns have been hashed into a primary key in the
        first ordinal position of the table. The hash uses the id_column_names
        so that only ID info is hashed into the hash value
        """
        # All ro
        output_file_obj = StringIO()
        for i, (pk_curr, row_curr) in enumerate(set(zip(pk, data))):
            new_line = [str(pk_curr)] + list(row_curr)  # type: ignore
            new_line = map(str, new_line)  # type: ignore
            new_line = map(  # type: ignore
                lambda x: x.replace("\t", "\\t")
                .replace("\n", "\\n")
                .replace("\r", "\\r")
                .replace("\\", "\\\\"),
                new_line,
            )
            output_file_obj.write("\t".join(new_line) + "\n")  # type: ignore

        output_file_obj.seek(0)

        return output_file_obj

    def _load(
        self,
        cxn: Conn,
        universe: "UNIVERSE_TYPE",
        rows: L[dict],
        insert: bool,
        test: bool = False,
    ) -> L[int]:
        """
        Helpful docstring
        """
        self._logger.debug("recursively loading foreign keys")
        for kk, vv in self.fks.items():
            if vv.insert:
                vv._load(cxn, universe, rows, insert=True, test=test)

        self._logger.debug("Getting attributes and generating hashes")
        obj_pk_name = universe[self.obj][0]
        pk, data = [], []
        for row in rows:
            pk_curr, data_curr = self._getvals(universe, row)
            pk.extend(pk_curr)
            data.extend(data_curr)

        self._logger.debug("writing data to stringio object")
        io_obj = self._data_to_stringIO(pk, data, obj_pk_name)
        if not data:
            return []

        # If we are testing don't try and do sql
        if not test:
            self._load_data(cxn, obj_pk_name, io_obj, insert)

        return [int(x) for x in pk]

    def _load_data(self, cxn: Conn, obj_pk_name: str, io_obj: StringIO, insert: bool) -> None:
        """
        Function that quickly loads an io_obj import the database specified
        obj_pk_name. Insert flag determines whether we update or insert.

        Args:
            cxn (Conn): connection to database to load intop
            obj_pk_name (str): name of objects primary key name
            io_obj (StringIO): StringIO to use for copy_from into database
            insert (bool): whether or not to update or insert into database

        Raises:
            ValueError: [description]
            DBgenExternalError: [description]
            ValueError: [description]
        """
        # Temporary table to copy data into
        # Set name to be hash of input rows to ensure uniqueness for parallelization
        temp_table_name = self.obj + "_temp_load_table_" + str(getrandbits(64))

        # Need to create a temp table to copy data into
        # Add an auto_inc column so that data can be ordered by its insert location

        create_temp_table = """
        DROP TABLE IF EXISTS {temp_table_name};
        CREATE TEMPORARY TABLE {temp_table_name} AS
        TABLE {obj}
        WITH NO DATA;
        ALTER TABLE {temp_table_name}
        ADD COLUMN auto_inc SERIAL NOT NULL;
        """.format(
            obj=self.obj, temp_table_name=temp_table_name
        )

        cols = [obj_pk_name] + list(sorted(self.attrs.keys())) + list(sorted(self.fks.keys()))
        from dbgen.templates import jinja_env

        if insert:
            template = jinja_env.get_template("insert.sql.jinja")
        else:
            template = jinja_env.get_template("update.sql.jinja")

        first = False
        update = True
        fk_cols = self.fks.keys()
        template_args = dict(
            obj=self.obj,
            obj_pk_name=obj_pk_name,
            partition_attr=self.partition_attr,
            temp_table_name=temp_table_name,
            all_column_names=cols,
            fk_cols=fk_cols,
            first=first,
            update=update,
        )
        load_statement = template.render(**template_args)

        with cxn.cursor() as curs:
            # Create the temp table
            curs.execute(create_temp_table)
            # Attempt the loading step 3 times
            self._logger.debug("load into temporary table")
            query_fail_count = 0
            while True:
                if query_fail_count == NUM_QUERY_TRIES:
                    raise ValueError("Query Cancel fail max reached")
                try:
                    curs.copy_from(
                        io_obj,
                        temp_table_name,
                        null="None",
                        columns=cols,
                    )
                    break
                except QueryCanceled:
                    print("Query cancel failed")
                    query_fail_count += 1
                    continue
                except Psycopg2Error as exc:
                    raise DBgenExternalError(exc.pgerror)

            # Try to insert everything from the temp table into the real table
            # If a foreign_key violation is hit, we delete those rows in the
            # temp table and move on
            fk_fail_count = 0
            self._logger.debug("transfer from temp table to main table")
            while True:
                if fk_fail_count == 10:
                    raise ValueError("User Canceled due to large number of FK violations")
                # check for ForeignKeyViolation error
                try:
                    curs.execute(load_statement)
                    break
                except psycopg2.errors.ForeignKeyViolation as exc:
                    pattern = r'Key \((\w+)\)=\(([\-\d]+)\) is not present in table "(\w+)"'
                    fk_name, fk_pk, fk_obj = re.findall(pattern, exc.pgerror)[0]
                    delete_statement = f"delete from {temp_table_name} where {fk_name} = {fk_pk}"
                    curs.execute(delete_statement)
                    self._logger.error(
                        "---\n"
                        f"ForeignKeyViolation #({fk_fail_count+1}): tried to insert {fk_pk} into"
                        f" FK column {fk_name} of {self.obj}."
                        f"\nBut no row exists with {fk_obj}_id = {fk_pk} in {fk_obj}."
                    )
                    self._logger.error(f"Moving on without inserting any rows with this {fk_pk}")
                    self._logger.error(exc)
                    fk_fail_count += 1
                    continue
            if fk_fail_count:
                self._logger.error(f"Fail count = {fk_fail_count}")

        io_obj.close()
        self._logger.debug("loading finished")

    def test(self, universe: "UNIVERSE_TYPE", rows: L[dict]) -> D[str, L[dict]]:
        """
        Takes in the universe and processed namespaces and generates dict where keys are table names and values are lists of input rows

        Args:
            universe (D[str, T[str, L[str], L[str]]]): universe of model
            rows (L[dict]): example processed namespaces after PyBlocks applied

        Returns:
            D[str, L[dict]]: dictionary of mapping tables to lists of rows that would be inserted if this load were called
        """
        obj_pk_name, ids, id_fks, dtype_dict = universe[self.obj]
        pk, data = [], []

        for row in rows:
            pk_curr, data_curr = self._getvals(universe, row)
            pk.extend(pk_curr)
            data.extend(data_curr)

            for kk, vv in sorted(self.fks.items()):
                if vv.pk is not None:
                    val = vv.pk.arg_get(row)
                else:
                    val, fk_adata = vv._getvals(universe, row)

        io_obj = self._data_to_stringIO(pk, data, obj_pk_name)

        cols = [obj_pk_name] + list(sorted(self.attrs.keys())) + list(sorted(self.fks.keys()))
        table_rows = []
        while True:
            line = io_obj.readline()
            if not line:
                break
            table_rows.append({col: val for col, val in zip(cols, line.strip("\n").split("\t"))})

        output = {self.obj + ("_insert" if self.insert else ""): table_rows}
        # Save the rows of recursive loads
        for kk, vv in self.fks.items():
            if vv.insert:
                if vv.obj == self.obj and vv.insert == self.insert:
                    print("!WARNING! self FKs aren't viewable in interact mode")
                else:
                    output.update(vv.test(universe, rows))
        return output

    def make_src(self) -> str:
        """
        Output a stringified version of load that can be run in an Airflow PythonOperator
        """
        attrs = ",".join([f"{k}={v.make_src(meta=True)}" for k, v in self.attrs.items()])
        template = (
            "Load(obj= '{{ obj }}',attrs= dict({{attrs}}),"
            "fks=dict({{ fks }}),pk= {{ pk }},insert={{ insert }})"
        )
        fks = ",".join([f"{k}={v.make_src()}" for k, v in self.fks.items()])
        pk = None if self.pk is None else self.pk.make_src(meta=True)
        return Template(template).render(obj=self.obj, attrs=attrs, fks=fks, pk=pk, insert=self.insert)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""parse_employees function"""
from csv import reader
from typing import Tuple as T


def parse_employees(pth: str) -> T[list, list, list, list, list]:
    """
    Takes in path to employees csv and returns a tuple of lists for each column

    Args:
        pth (str): path to employee csv

    Returns:
        T[list, list, list, list, list]: tuple of lists (employees, salary,
        managers, department, secretary)
    """
    with open(pth) as f:
        r = reader(f)
        next(r)
        ename, sal, man, dname, sec = tuple(map(list, zip(*r)))  # type: ignore
    return ename, [float(s) for s in sal], man, dname, sec  # type: ignore



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import readline
from functools import reduce
from os import system
from typing import TYPE_CHECKING, Any
from typing import Callable as C
from typing import Dict as D
from typing import List as L
from typing import Optional as Opt
from typing import Set as S
from typing import Tuple as T

if TYPE_CHECKING:
    from dbgen.core.gen import Generator


def interact_gen(
    objs, gen: "Generator", input_rows: L[dict], max_rows: int = 200
) -> T[L[D[str, dict]], L[D[str, L[dict]]]]:
    """
    Allows a CLI for interacting with a generator given a set of input rows

    Args:
        gen (Gen): generator to interact with
        input_rows (L[dict]): list of input rows that simulate the query input

    Raises:
        ImportError: if prettytable, pprint and readline not installed (didn't want to require these in the requirements.txt just for this function)

    Returns:
        L[dict]: output of the generator
    """
    # Attempt to import necessary functions for formmatting
    try:
        from pprint import pprint

        from prettytable import PrettyTable  # type: ignore
    except ImportError:
        raise ImportError("Need pprint and prettytable for interact mode")

    # Initialize output list and valid responses for the outer question
    post_pyblocks: L[dict] = []
    load_dicts: L[dict] = []
    valid_responses = ("s", "c", "q", "m")
    # Initialize the response to user input and formatting delimiter func
    answer: Opt[str] = ""
    delimiter = lambda: print("#----------------------------------------------------------------------------")
    # Pretty table for displaying input row data
    if input_rows:
        x = PrettyTable(field_names=list(input_rows[0].keys()))
    else:
        print("Generator has no inputs!")

    while len(input_rows) > 0 and answer != "q":
        print_len = 200
        next_row = dict(input_rows.pop(0))
        next_row_str = lambda x: (str(next_row)[:x] + "..." if len(str(next_row)) > x else dict(next_row))
        system("clear")
        print("Next Row:")
        x.add_row(next_row.values())
        pprint(next_row_str(print_len))
        answer = None
        while answer not in (
            "c",
            "s",
        ):
            if answer == "m":
                print_len += 200
                pprint(next_row_str(print_len))
            elif answer not in valid_responses and answer is not None:
                delimiter()
                print("Invalid response, please select s/c/q/m")
                delimiter()
            else:
                delimiter()
            answer = input("Continue(c), quit (q), skip (s), or expand current row(m)?\n").lower()
            if answer == "q":
                print("Quitting...")
                return post_pyblocks, load_dicts
        # Skip the row and clear the pretty table
        if answer == "s":
            x.clear_rows()
            continue
        delimiter()
        print("Processing Row...")
        list_of_processed_namespaces, curr_load_dicts = gen.test(objs, [next_row], verbose=True)
        curr_output = list_of_processed_namespaces[0]
        curr_load_dict = curr_load_dicts[0]
        system("clear")
        print("Next Row:")
        pprint(next_row_str(print_len))

        completer = get_completer(list(curr_output.keys()))
        readline.parse_and_bind("tab: complete")
        readline.set_completer(completer)
        display = ""
        delimiter()
        while display != "q":
            print("PyBlock Names:")
            for key in curr_output.keys():
                print("\t- " + key)
            delimiter()
            display = input("What pyblock to display? (tab to see options/q to quit)\n").lower()
            if display in curr_output:
                delimiter()
                if display == "query":
                    print("Query Row:")
                else:
                    print(f"Function Name: {display}")
                print(f"Output:")
                try:
                    pprint(curr_output[display])
                except KeyboardInterrupt:
                    pass
                delimiter()
                while True:
                    if input("Press enter to continue\n") == "":
                        system("clear")
                        break
            elif display == "":
                system("clear")
                continue
            else:
                print("PyBlock Names:\n")
                pprint(list(curr_output.keys()))
                delimiter()
                print(f"Key not found in processed dict: {display}")
                delimiter()

        completer = get_completer(list(curr_load_dict.keys()))
        readline.set_completer(completer)
        display = ""
        while display != "q":
            print("Load Names:")
            for key in curr_load_dict.keys():
                print("\t- " + key + f" ({len(curr_load_dict[key])} rows)")

            delimiter()
            display = input("What load to display? (tab to see options/q to quit)\n").lower()
            if display in curr_load_dict:
                delimiter()
                print(f"Table Name: {display}")
                print(f"Output:")
                try:
                    rows = curr_load_dict[display]
                    if rows:
                        all_keys: S[str] = reduce(
                            lambda prev, next: prev.union(set(next.keys())),
                            rows,
                            set(),
                        )
                        table = PrettyTable(field_names=list(all_keys))
                        for row in curr_load_dict[display][:max_rows]:
                            row_data = list(row.get(key) for key in all_keys)
                            row_data = [(data if data != "" else '""') for data in row_data]
                            table.add_row(row_data)
                        print(table)
                    else:
                        print("No rows to be updated or inserted")
                except KeyboardInterrupt:
                    pass
                delimiter()
                while True:
                    if input("Press enter to continue\n") == "":
                        system("clear")
                        break
            elif display == "":
                system("clear")
                continue
            else:
                delimiter()
                print(f"Key not found in Load Dict: {display}")
                delimiter()

        post_pyblocks.append(curr_output)
        load_dicts.append(curr_load_dict)
        x.clear_rows()

    print("No more rows left")
    return post_pyblocks, load_dicts


def get_completer(cmds: L[Any]) -> C[[str, int], Any]:
    def completer(text, state):
        options = [i for i in cmds if i.startswith(text)]
        if state < len(options):
            return options[state]
        else:
            return None

    return completer



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Welcome to dbgen"""

from dbgen.version import git_version, version

__author__ = "Modelyst LLC"
__email__ = "info@modelyst.io"
__maintainer__ = "Michael Statt"
__maintainer_email__ = "michael.statt@modelyst.io"
__version__ = version
__gitversion__ = git_version
# Build Logo for easy CLI version checking
git_ver_str = f"GITVERSION: {__gitversion__}\n" if __gitversion__ else ""
LOGO = f"""
-------------------------------
    ____  ____
   / __ \\/ __ )____ ____  ____
  / / / / __  / __ `/ _ \\/ __ \\
 / /_/ / /_/ / /_/ /  __/ / / /
/_____/_____/\\__, /\\___/_/ /_/
            /____/
-------------------------------
VERSION: {__version__}
{git_ver_str}-------------------------------
"""

# Imports
from dbgen.core.expr.expr import (
    ABS,
    AND,
    ARRAY,
    ARRAY_AGG,
    AVG,
    BINARY,
    CASE,
    COALESCE,
    CONCAT,
    CONVERT,
    COUNT,
    EQ,
    GE,
    GROUP_CONCAT,
    GT,
    IF_ELSE,
    IN,
    JSON_EXTRACT,
    LE,
    LEFT,
    LEN,
    LIKE,
    LT,
    MAX,
    MIN,
    NE,
    NOT,
    NULL,
    OR,
    POSITION,
    R2,
    REGEXP,
    REPLACE,
    RIGHT,
    STD,
    SUBSELECT,
    SUM,
    Expr,
    Literal,
    One,
    Sum,
    Tup,
    Zero,
    false,
    toDecimal,
    true,
)
from dbgen.core.expr.pathattr import PathAttr
from dbgen.core.expr.sqltypes import (
    JSON,
    JSONB,
    Boolean,
    Date,
    Decimal,
    Double,
    Int,
    SQLType,
    Text,
    Timestamp,
    Varchar,
)
from dbgen.core.fromclause import Path as JPath
from dbgen.core.func import Env, Func, Import
from dbgen.core.funclike import Arg, Const, PyBlock
from dbgen.core.gen import Generator
from dbgen.core.gen import Generator as Gen
from dbgen.core.misc import ConnectInfo
from dbgen.core.model.model import Model
from dbgen.core.query import Query
from dbgen.core.schema import Attr, Entity, Path, PathEQ, QView, RawView
from dbgen.core.schema import UserRel as Rel
from dbgen.utils.lists import concat_map, flatten, merge_dicts, nub
from dbgen.utils.parsing import parser

############################
from dbgen.utils.sql import sqlexecute, sqlselect



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External modules
from os.path import join

# Internal Modules
from dbgen import Const, Env, Generator, Import, Model, PyBlock

from ..scripts.io.parse_employees import parse_employees  # type: ignore
from ..utils import get_config  # type: ignore

################################################################################


def io(mod: Model) -> None:
    config = get_config()

    # Extract tables
    tabs = ["employee", "department"]

    Emp, Dept = map(mod.get, tabs)

    ################################################################################

    pe_func = PyBlock(
        parse_employees,
        env=Env([Import("csv", "reader")]),
        args=[Const(join(config["dbgen"]["model_root"], "data/example.csv"))],
        outnames=["ename", "sal", "man", "dname", "sec"],
    )

    pop_employees = Generator(
        name="pop_employees",
        desc="parses CSV file with employee info",
        transforms=[pe_func],
        tags=["io"],
        loads=[
            Emp(
                insert=True,
                name=pe_func["ename"],
                salary=pe_func["sal"],
                manager=Emp(insert=True, name=pe_func["man"]),
                department=Dept(
                    insert=True,
                    name=pe_func["dname"],
                ),
            )
        ],
    )

    ################################################################################
    gens = [pop_employees]
    mod.add(gens)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import ast
from importlib.util import module_from_spec, spec_from_file_location
from inspect import (
    getdoc,
    getmembers,
    getsourcefile,
    getsourcelines,
    isbuiltin,
    isclass,
    isfunction,
    signature,
)
from os.path import exists
from pathlib import Path
from re import findall
from sys import version_info

# External Modules
from typing import Any
from typing import Callable as C
from typing import Dict as D
from typing import List as L
from typing import Union as U

from hypothesis.strategies import SearchStrategy, builds

# Iternal Modules
from dbgen.core.datatypes import DataType, Tuple
from dbgen.utils.config import DBGEN_TMP
from dbgen.utils.exceptions import DBgenInternalError, DBgenInvalidArgument
from dbgen.utils.misc import Base, hash_
from dbgen.utils.sql import Connection as Conn
from dbgen.utils.sql import mkInsCmd, mkSelectCmd, sqlexecute, sqlselect

"""
Defines the Func class, which is initialized with any Python callable but gets
enriched in the __init__ with a lot more information and methods (from inspect)
"""
assert version_info[1] > 5, "Stop using old python3 (need 3.x, x > 5)"
################################################################################
class Import(Base):
    """
    Representation of an Python import line.

    Examples:

    --> from libname import unaliased, things, aliased as Thing
        Import('libname',         <---- (DO NOT make this a keyword argument!!!)
                'unaliased',
                'things',
                aliased = 'Thing')

    --> import numpy as np
        Import('numpy', alias = 'np')

    Unsupported edge case: trying to import some variable literally named
                            "alias" using an alias
    """

    def __init__(
        self,
        lib: str,
        unaliased_imports: U[L[str], str] = None,
        lib_alias: str = "",
        aliased_imports: D[str, str] = None,
    ) -> None:

        err = "Can't import %s as %s AND import specific terms (%s,%s) at once"
        terms = unaliased_imports or aliased_imports
        assert not (lib_alias and terms), err % (
            lib,
            lib_alias,
            unaliased_imports,
            aliased_imports,
        )

        self.lib = lib
        self.lib_alias = lib_alias

        if isinstance(unaliased_imports, str):
            self.unaliased_imports = [unaliased_imports]
        else:
            self.unaliased_imports = unaliased_imports or []
        self.aliased_imports = aliased_imports or {}
        super().__init__()

    def __str__(self) -> str:
        if not (self.unaliased_imports or self.aliased_imports):
            alias = f"as {self.lib_alias}" if self.lib_alias else ""
            return f"import {self.lib} {alias}"
        else:
            als = [f"{k} as {v}" for k, v in self.aliased_imports.items()]
            terms = list(self.unaliased_imports) + als
            return f"from {self.lib} import {', '.join(terms)}"

    def __eq__(self, other: object) -> bool:
        return False if not isinstance(other, Import) else vars(self) == vars(other)

    def __hash__(self) -> int:
        return int(self.hash)

    def __lt__(self, other: Any) -> bool:
        if type(self) == type(other):
            return repr(self) < repr(other)
        raise TypeError(f"'<' not supported between instances of '{type(self)}' and '{type(other)}'")

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)

    @staticmethod
    def from_str(s: str) -> "Import":
        """Parse a header line (parens not supported yet)"""
        if s[:6] == "import":
            if " as " in s[7:]:
                lib, lib_alias = [s[7:].split("as")[i].strip() for i in range(2)]
                return Import(lib, lib_alias=lib_alias)
            return Import(s.split()[1].strip())
        else:
            i = s.find("import")
            a, lib = s[:i].split()
            assert a == "from", "Bad source code beginning : \n\n\n" + s
            pat = r"([a-zA-Z0-9\_]+\s*(?:as\s*[a-zA-Z0-9\_]+)?)"
            groups = findall(pat, s[i + 6 :])
            objs = [list(map(str.strip, g.split("as"))) for g in groups]
            # objs_ = [x[0] if len(x) == 1 else tuple(x) for x in objs]
            unalias = [x[0] for x in objs if len(x) == 1]
            aliased = {x[0]: x[1] for x in objs if len(x) == 2}

            return Import(lib, unaliased_imports=unalias, aliased_imports=aliased)


class Env(Base):
    """
    Environment in which a python statement gets executed
    """

    def __init__(self, imports: L[Import] = None) -> None:

        if imports:
            assert isinstance(
                imports, list
            ), "Env takes in a list of imports. If there is 1 import wrap it in a list"

            self.imports = sorted(imports)
        else:
            self.imports = []
        super().__init__()

    def __str__(self) -> str:
        return "\n".join(map(str, self.imports))

    def __add__(self, other: "Env") -> "Env":
        return Env(list(set(self.imports + other.imports)))

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)

    # Public methods #

    @staticmethod
    def from_str(import_string: str) -> "Env":
        """Parse a header"""
        imports = []
        for node in ast.iter_child_nodes(ast.parse(import_string)):
            if isinstance(node, ast.ImportFrom):
                lib = node.module
                assert lib, f"Empty lib found {lib}\n{import_string}"
                aliased_imports = {}
                unaliased_imports = []
                for module in node.names:
                    if module.asname:
                        aliased_imports[module.name] = module.asname
                    else:
                        unaliased_imports.append(module.name)
                imports.append(
                    Import(
                        lib,
                        unaliased_imports=unaliased_imports,
                        aliased_imports=aliased_imports,
                    )
                )
            elif isinstance(node, ast.Import):
                assert len(node.names) == 1, f"Bad Import String! {import_string}"
                lib = node.names[0].name
                lib_alias = node.names[0].asname or ""
                imports.append(Import(lib, lib_alias=lib_alias))
        return Env(imports)

    @classmethod
    def from_file(cls, pth: Path) -> "Env":
        with open(pth) as f:
            return cls.from_str(f.read())


################################################################################
class Func(Base):
    """
    A function that can be used during the DB generation process.
    """

    def __init__(self, src: str, env: Env) -> None:
        assert isinstance(src, str), f"Expected src str, but got {type(src)}"
        assert isinstance(env, Env), f"Expected Env, but got {type(env)}"

        self.src = src
        self.env = env
        super().__init__()

    def __str__(self) -> str:
        n = self.src.count("\n")
        s = "" if n == 1 else "s"
        return "<Func (%d line%s)>" % (n, s)

    def __call__(self, *args: Any) -> Any:
        if hasattr(self, "_func") and self.path.exists():
            return self._func(*args)
        else:
            f = self._from_src()
            return f(*args)

    def __repr__(self) -> str:
        return self.name

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)

    # Properties #

    @property
    def name(self) -> str:
        return self._from_src().__name__

    @property
    def is_lam(self) -> bool:
        return self.src[:6] == "lambda"

    @property
    def doc(self) -> str:
        return getdoc(self._from_src()) or ""

    @property
    def sig(self) -> Any:
        return signature(self._from_src())

    @property
    def argnames(self) -> L[str]:
        return list(self.sig.parameters)

    @property
    def nIn(self) -> int:
        return len(self.inTypes)

    @property
    def notImp(self) -> bool:
        return "NotImplementedError" in self.src

    @property
    def output(self) -> Any:
        return self.sig.return_annotation

    @property
    def nOut(self) -> int:
        return len(self.outTypes)

    @property
    def inTypes(self) -> L[DataType]:
        return [DataType.get_datatype(x.annotation) for x in self.sig.parameters.values()]

    @property
    def path(self) -> Path:
        return DBGEN_TMP / (str(hash_(self.file())) + ".py")

    @property
    def outTypes(self) -> L[DataType]:
        ot = DataType.get_datatype(self.output)
        if len(ot) == 1:
            return [ot]
        else:
            assert isinstance(ot, Tuple)
            return ot.args

    def file(self) -> str:
        lam = "f = " if self.is_lam else ""
        return str(self.env) + "\n" + lam + self.src

    # Private methods #

    def _from_src(self, force: bool = False) -> C:
        """
        Execute source code to get a callable
        """

        if force or not exists(self.path):
            with open(self.path, "w") as t:
                t.write(self.file())

        f = self.path_to_func(str(self.path))

        return f

    # Public methods #
    def store_func(self, force: bool = False) -> None:
        """Load func from source code and store as attribute (better performance
        but object is no longer serializable / comparable for equality )
        """
        self._func = self._from_src(force)

    def del_func(self) -> None:
        """Remove callable attribute after performance is no longer needed"""
        if hasattr(self, "_func"):
            del self._func

    def add(self, cxn: Conn) -> int:
        """
        Log function data to metaDB, return its ID
        """
        q = mkSelectCmd("_func", ["func_id"], ["checksum"])
        f_id = sqlselect(cxn, q, [hash_(self.src)])
        if f_id:
            return f_id[0][0]
        else:

            cols = [
                "name",
                "checksum",
                "source",
                "docstring",
                "inTypes",
                "outType",
                "n_in",
                "n_out",
            ]

            binds = [
                self.name,
                hash_(self.src),
                self.src,
                self.doc,
                str(self.inTypes),
                str(self.outTypes),
                self.nIn,
                self.nOut,
            ]

            sqlexecute(cxn, mkInsCmd("_func", cols), binds)
            f_id = sqlselect(cxn, q, [hash_(self.src)])
            return f_id[0][0]

    @staticmethod
    def path_to_func(pth: str) -> C:

        try:
            spec = spec_from_file_location("random", pth)
            mod = module_from_spec(spec)
            assert spec and spec.loader, "Spec or Spec.loader are broken"
            spec.loader.exec_module(mod)  # type: ignore
            transforms = [o for o in getmembers(mod) if isfunction(o[1]) and getsourcefile(o[1]) == pth]
            assert len(transforms) == 1, "Bad input file %s has %d functions, not 1" % (
                pth,
                len(transforms),
            )
            return transforms[0][1]

        except Exception as e:
            if exists(pth):
                with open(pth) as f:
                    content = f.read()
            raise DBgenInternalError(
                f"Error while trying to load source code. You may be missing an import in your PyBlocks Env object. \nPath:{pth}\nFile Contents:\n--------\n{content}\n--------\nLoad Error: {e}"
            )

    @classmethod
    def from_callable(cls, f: U[C, "Func"], env: Env = Env()) -> "Func":
        """
        Generate a func from a variety of possible input data types.
        """
        if isinstance(f, Func):
            # assert not getattr(env,'imports',False)
            return f
        else:
            assert callable(f), f"tried to instantiate Func, but not callable {type(f)}"
            return Func(src=cls.get_source(f), env=env)

    @staticmethod
    def get_source(f: C) -> str:
        """
        Return the source code, even if it's lambda function.
        """
        # Check for built in functions as their source code can not be fetched
        if isbuiltin(f) or (isclass(f) and getattr(f, "__module__", "") == "builtins"):
            raise DBgenInvalidArgument(
                f"Error getting source code for PyBlock. {f} is a built-in function.\n"
                f"Please wrap in lambda like so: `lambda x: {f.__name__}(x)`"
            )

        try:
            source_lines, _ = getsourcelines(f)
        except (OSError, TypeError) as e:
            # functions defined in pdb / REPL / eval / some other way in which
            # source code not clear
            raise ValueError("from_callable: ", f, e)

        # Handle 'def'-ed functions and long lambdas
        src = "".join(source_lines).strip()

        if len(source_lines) > 1 and src[:3] == "def":
            return src

        err = 'Only one "lambda" allowed per line: '
        assert src.count("lambda") == 1, err + src

        src_ = src[src.find("lambda") :]  # start of lambda function

        # Slice off trailing chars until we get a callable function
        while len(src_) > 6:
            try:
                if callable(eval(src_)):
                    return src_
            except (SyntaxError, NameError):
                pass

            src_ = src_[:-1].strip()

        raise ValueError("could not parse lambda: " + src)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
import logging
from dataclasses import asdict, replace
from enum import Enum
from pathlib import Path
from types import FunctionType
from typing import List, Optional

import typer

from dbgen import LOGO
from dbgen.core.model.model import Model
from dbgen.utils import settings
from dbgen.utils.config import RunConfig, config

# Errors
ERROR_FORMAT = "Model is not in MODULE:PACKAGE format: {0}"
ERROR_MODULE = "Could not find module:\nModule: {0}\nPackage: {1}\nError: {2}"
ERROR_PACKAGE = "Could not find package within module:\nModule: {0}\nPackage: {1}\nError: {2}"
ERROR_NOT_MODEL = "Import String is not for a DBgen Model: \nImport String: {0}\nClass: {1}"
ERROR_NOT_MODEL_FUNCTION = "Import String is for a function that does not produce a DBgen Model: \nImport String: {0}\nOutput Class: {1}"
ERROR_RUNNING_MODEL_FACT = "Import String is for a function produced an error or required arguments: \nImport String: {0}\nOutput Class: {1}"

app = typer.Typer(help="DBgen Model Runner")
state = {"confirm": True}
logger = logging.getLogger(__name__)
LOG_MAP = {
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "WARNING": logging.WARNING,
    "CRITICAL": logging.CRITICAL,
}


class LogLevel(str, Enum):
    """enum for setting logging level on CLI"""

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    CRITICAL = "CRITICAL"


def confirm_nuke(value: bool):
    """
    Confirm that user wants to nuke database

    Args:
        value (bool): nuke_value
    """
    if value and state["confirm"]:
        confirm = input("Are you sure you want to nuke the database? (Y/n) ")
        if confirm.lower() != "y":
            exit(1)
    return value


def set_confirm(value: bool):
    """Auto confirm all prompted values"""
    state["confirm"] = not value


def version_callback(value: bool):
    """
    Eagerly print the version LOGO

    Args:
        value (bool): [description]

    Raises:
        typer.Exit: exits after showing version
    """
    if value:
        typer.echo(LOGO)
        raise typer.Exit()


def file_existence(value: Path):
    if value and value.exists():
        return value
    elif value is not None:
        # No config found warn user
        logger.warning(f"Provided config {value.absolute()} does not exist.")
    return None


def validate_model_str(model_str: str) -> Model:
    """
    Validate the user input model import str checking for malformed and invalid inputs.py

    Args:
        model_str (str): CLI/config file input in MODULE:PACKAGE format

    Raises:
        typer.BadParameter: Whenever one of many checks is failed

    Returns:
        Model: the output model after checks are in place
    """
    basic_error = lambda fmt, val: typer.BadParameter(fmt.format(*val))

    # Check for empty string model_strs
    if model_str in (":", ""):
        raise basic_error(ERROR_FORMAT, [model_str])

    split_model = model_str.split(":")
    if len(split_model) != 2:
        raise basic_error(ERROR_FORMAT, [model_str])
    module, package = split_model
    try:
        _temp = __import__(module, globals(), locals(), [package])
        model = getattr(_temp, package)
        if isinstance(model, Model):
            return model
        elif isinstance(model, FunctionType):
            try:
                model = model()
            except TypeError:
                raise basic_error(ERROR_RUNNING_MODEL_FACT, [model_str, type(model).__name__])
            if isinstance(model, Model):
                return model
            raise basic_error(ERROR_NOT_MODEL_FUNCTION, [model_str, type(model).__name__])

        raise basic_error(ERROR_NOT_MODEL, [model_str, type(model).__name__])
    except ModuleNotFoundError as exc:
        if "No module" in str(exc):
            raise basic_error(ERROR_MODULE, [module, package, str(exc)])
        raise basic_error(ERROR_PACKAGE, [module, package, str(exc)])
    except AttributeError as exc:
        raise typer.BadParameter(str(exc))
    return model


@app.command()
def run(
    model_str: str = typer.Argument(
        ...,
        help="An import string in MODULE:PACKAGE format where the package is either a dbgen model variable or a function that produces one",
    ),
    only: List[str] = typer.Option([], help="Generators to include"),
    xclude: List[str] = typer.Option([], help="Generators to xclude"),
    add: bool = typer.Option(None, help="Add the database objects (tables/columns) before run."),
    retry: bool = typer.Option(None, help="Ignore repeat checking"),
    start: Optional[str] = typer.Option(None, help="Generator to start run at"),
    until: Optional[str] = typer.Option(None, help="Generator to finish run at."),
    serial: bool = None,
    config_file: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        help="DBgen config file to use for specifying run parameters as well as DB",
        envvar="DBGEN_CONFIG",
        callback=file_existence,
    ),
    nuke: bool = typer.Option(
        None,
        help="Delete the entire db and meta schema.",
        callback=confirm_nuke,
    ),
    no_conf: bool = typer.Option(
        False,
        "--no-confirm",
        "-y",
        is_eager=True,
        callback=set_confirm,
    ),
    bar: bool = typer.Option(
        None,
        help="Show tqdm progress bar for the run. Best to disable for Airflow Runs",
    ),
    skip_row_count: bool = typer.Option(False, help="Skip the row count"),
    batch: int = typer.Option(None, help="Batch size for the run. Overrides any Gen-level batch size"),
    write_logs: bool = typer.Option(None, help="Write the logs for the dbgen to file."),
    log_level_str: LogLevel = typer.Option(
        LogLevel.INFO, "--log-level", "-L", help="Set the level of logging"
    ),
    log_path: Path = typer.Option(
        None,
        help="Location for log file, overrides the default of $HOME/.dbgen/dbgen.log.",
    ),
    print_logo: bool = False,
    version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True),
) -> int:
    """
    Run a dbgen model from command line.py

    Args:
        model (Model): [description].
        only (List[str], optional): [description].
        xclude (List[str], optional): [description].
        add (bool, optional): [description].
        retry (bool, optional): [description].
        start (str, optional): [description].
        until (str, optional): [description].
        serial (bool, optional): [description].
        config_file (Optional[Path], optional): [description].
        nuke (bool, optional): [description].
        no_conf (bool, optional): [description].
        bar (bool, optional): [description].
        skip_row_count (bool, optional): [description].
        batch (int, optional): [description].
        write_logs (bool, optional): [description].
        log_level_str (LogLevel, optional): [description].
        log_path (str, optional): Location for log file, overrides the default of $HOME/.dbgen/dbgen.log
        print_logo (bool, optional): [description].
        version (bool, optional): [description].

    Returns:
        dict: [description]
    """
    if print_logo:
        print(LOGO)
    # Parse inputs
    run_config = RunConfig()
    prune = lambda d: {k: v for k, v in d.items() if v is not None and k in run_config.fields}
    prune_inputs = prune(vars())

    # Update the config with the cmd line config
    if config_file:
        config.read(config_file)
    # Use the config file to set RunConfig values
    run_config = replace(run_config, **(prune(config.getsection("run")) or {}))
    run_config = replace(run_config, **(prune(config.getsection("logging")) or {}))
    # Use the CLI to override any configs
    run_config = replace(run_config, **prune_inputs)
    # Initialize settings to get the connections
    settings.initialize()
    model = validate_model_str(model_str)

    # Validate gen and tag related inputs
    all_names = list(only) + list(xclude)
    all_names += [start] if start else []
    all_names += [until] if until else []
    for val in all_names:
        try:
            model._validate_name(val)
        except ValueError as e:
            if val in only:
                param = "only"
            elif val in xclude:
                param = "xclude"
            elif val == start:
                param = "start"
            else:
                param = "until"
            raise typer.BadParameter(f"{param.upper()} - {e}")
    # Process Log-Level
    assert settings.CONN and settings.META_CONN
    model.run(settings.CONN, settings.META_CONN, **asdict(run_config))
    return 0


@app.command()
def serialize(
    model_str: str = typer.Argument(
        ...,
        help="An import string in MODULE:PACKAGE format where the package is either a dbgen model variable or a function that produces one",
    ),
    out_pth: Path = typer.Argument(
        Path("./model.json"),
        help="Output path to write the model.json to",
    ),
):
    """
    Serializes the DBgen model into a json
    """
    model = validate_model_str(model_str)
    with open(out_pth, "w") as f:
        f.write(model.toJSON())
    typer.echo(f"Finished! Your model is serialized at {out_pth.absolute()}")
    typer.echo(f"Hash: {model.hash}")


@app.command()
def deserialize(
    model_path: Path = typer.Argument(
        ...,
        help="Location of stored model.json file to read in",
    ),
):
    """
    Reads a model.json to check for model errors
    """
    try:
        model = Model.fromJSON(model_path.read_text())
    except AssertionError:
        raise typer.BadParameter("Invalid json")
    except json.decoder.JSONDecodeError:
        raise typer.BadParameter("Invalid json")
    assert isinstance(model, Model)
    typer.echo(f"Parsed Model: {model}")
    typer.echo(f"Hash: {model.hash}")
    with open("test.json", "w") as f:
        f.write(model.toJSON())


@app.command()
def airflow(
    model_str: str = typer.Argument(
        ...,
        help="An import string in MODULE:PACKAGE format where the package is either a dbgen model variable or a function that produces one",
    ),
    out_pth: Path = typer.Option(
        None,
        help="Output path to write the model.json to",
    ),
):
    """
    Serializes the DBgen model into a json
    """
    model = validate_model_str(model_str)
    if not out_pth:
        out_pth = Path(f"./{model.name}.py")
    with open(out_pth, "w") as f:
        f.write(model.run_airflow())
    typer.echo(f"Finished! Your model is serialized at {out_pth.absolute()}")
    typer.echo(f"Hash: {model.hash}")


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dbgen',
 'dbgen.cli',
 'dbgen.cli.newfiles',
 'dbgen.core',
 'dbgen.core.airflow_plugin',
 'dbgen.core.expr',
 'dbgen.core.model',
 'dbgen.example',
 'dbgen.example.generators',
 'dbgen.example.scripts',
 'dbgen.templates',
 'dbgen.utils']

package_data = \
{'': ['*'], 'dbgen.example': ['data/*'], 'dbgen.utils': ['config_templates/*']}

install_requires = \
['hypothesis>=5.23.7',
 'infinite>=0.1',
 'jinja2>=2.10.0',
 'networkx>=2.4',
 'pathos>=0.2.6',
 'psycopg2-binary>=2.7.4',
 'sshtunnel>=0.1.5,<0.2.0',
 'tqdm>=4.48.0',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{':extra == "docs"': ['pdocs>=1.1.1,<2.0.0'],
 'airflow': ['apache-airflow[postgres]>=2.1.0,<3.0.0'],
 'docs': ['markdown-include==0.6.0',
          'mkdocs>=1.1.2,<2.0.0',
          'mkdocs-autorefs>=0.1.1,<0.2.0',
          'mkdocs-markdownextradata-plugin>=0.2.4,<0.3.0',
          'mkdocs-material>=7.0.6,<8.0.0',
          'mkdocstrings>=0.15.0,<0.16.0']}

entry_points = \
{'console_scripts': ['dbgen = dbgen.__main__:main']}

setup_kwargs = {
    'name': 'modelyst-dbgen',
    'version': '0.4.6',
    'description': 'DBgen (Database Generator) is an open-source Python library for connecting raw data, scientific theories, and relational databases',
    'long_description': '<!--\n   Copyright 2021 Modelyst LLC\n\n   Licensed under the Apache License, Version 2.0 (the "License");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an "AS IS" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.\n -->\n\n# DBgen\n\n<p align="center">\n  <a href="https://dbgen.modelyst.com"><img src="docs/img/dbgen_logo.png" alt="DBgen"></a>\n</p>\n---\n\n**Documentation**: <a href="https://dbgen.modelyst.com" target="_blank">https://dbgen.modelyst.com</a>\n\n**Github**: <a href="https://github.com/modelyst/dbgen" target="_blank">https://github.com/modelyst/dbgen</a>\n\n---\n\nDBgen (Database Generator) is an open-source Python library for\nconnecting raw data, scientific theories, and relational databases.\nThese are some of the main features:\n\n1.  Very easy to work with\n2.  Integration with the PostgreSQL databases.\n\nDBgen was initially developed by [Modelyst](https://www.modelyst.com/).\n\n## What is DBgen?\n\nDBgen was designed to support scientific data analysis with the following\ncharacteristics:\n\n1.  Transparent\n\n    - Because scientific efforts ought be shareable and mutually\n      understandable.\n\n2.  Flexible\n\n    - Because scientific theories are under continuous flux.\n\n3.  Maintainable\n    - Because the underlying scientific models one works with are\n      complicated enough on their own, we can\'t afford to introduce\n      any more complexity via our framework.\n\nDBGen is an opinionated ETL tool. ETL tools exist but they rarely\ngive the tools necessary for a scientific workflow. Opinionated\naspect: it really cares about what the end product is (ID columns on\nall the tables). We\'re dealing with a much more restricted ETL\nproblem (extracting and ).\n\nComparison to\n\n1. [Airflow](https://airflow.apache.org/)\n\n   - Has a priority for ETL scalability\n\n2. [Fireworks](https://materialsproject.github.io/fireworks/)\n\n3. [AiiDA](http://www.aiida.net/) or [Atomate](https://atomate.org/)\n   - We don\'t focus on the actual submission of computational\n     science workflows.\n\n## What isn\'t DBgen?\n\n1. An [ORM](https://en.wikipedia.org/wiki/Object-relational_mapping) tool (see [Hibernate](http://hibernate.org/orm/) for Java or [SQLAlchemy](https://www.sqlalchemy.org/) for Python)\n\n   - DBgen operates at a higher level of abstrload, not exposing the user to low level SQL commands like SELECT or INSERT.\n\n2. A database manager (see\n   [MySQLWorkbench](https://www.mysql.com/products/workbench/),\n   [DBeaver](https://dbeaver.io/), [TablePlus](https://tableplus.com/),\n   etc.)\n3. An opinioniated tool with a particular schema for scientific data /\n   theories.\n\n## Getting DBgen\n\n### Via Github\n\nCurrently, the only method of installing DBgen is through Github. This is best done by using the [poetry](https://python-poetry.org/) package manager. To do this, first clone the repo to a local directory. Then use the command `poetry install` in the directory to install the required dependencies. You will need at least python 3.8 to install the package.\n\nThis installation process is still in development, please contact the DBgen if you encounter issues. Moving forward we plan to use make dbgen installable through `pip`.\n\n### API documentation\n\nDocumentation of modules and classes can be found in\nAPI docs \\</modules\\>.\n\n#### Reporting bugs\n\nPlease report any bugs and issues at DBgen\'s [Github Issues\npage](https://github.com/modelyst/dbgen/issues).\n\n## License\n\nDBgen is released under the [Apache 2.0 License](license/).\n',
    'author': 'Michael Statt',
    'author_email': 'michael.statt@modelyst.io',
    'maintainer': 'Michael Statt',
    'maintainer_email': 'michael.statt@modelyst.io',
    'url': 'https://www.dbgen.modelyst.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from pprint import pformat

# External Modules
from typing import List

from networkx import DiGraph, NetworkXUnfeasible  # type: ignore
from networkx.algorithms import lexicographical_topological_sort, simple_cycles  # type: ignore


# Graph
# --------
def topsort_with_dict(G: DiGraph, d: dict) -> List:
    """
    Assuming a graph with object names and dict mapping names to objects,
    perform a topsort and return the list of objects.
    """
    try:
        sortd = list(lexicographical_topological_sort(G))
        return [d[x] for x in sortd]
    except NetworkXUnfeasible:
        cycles = pformat(list(simple_cycles(G)))
        raise ValueError(f"Cycles found: {cycles}")


def make_acyclic(G: DiGraph) -> DiGraph:
    stack = list(simple_cycles(G))
    while stack:
        cyc = stack.pop()
        G.remove_edge(cyc[0], cyc[1])  # this is bad, remove random edge
        stack = list(simple_cycles(G))
    return G



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractmethod

# External
from typing import TYPE_CHECKING, Any
from typing import Dict as D
from typing import Iterator as Iter
from typing import List as L
from typing import Sequence
from typing import Set as S
from typing import Tuple as T

from hypothesis.strategies import SearchStrategy, builds, lists

from dbgen.core.expr.expr import PK
from dbgen.core.expr.sqltypes import Int, SQLType, Text, Varchar
from dbgen.core.funclike import Arg, ArgLike, Const
from dbgen.core.load import Load
from dbgen.core.misc import Dep
from dbgen.utils.exceptions import DBgenInvalidArgument, DBgenMissingInfo
from dbgen.utils.misc import Base
from dbgen.utils.sql import Connection as Conn
from dbgen.utils.sql import mkInsCmd, mkSelectCmd, sqlexecute, sqlselect

# Internal
if TYPE_CHECKING:
    from dbgen.core.expr.pathattr import PathAttr
    from dbgen.core.model.model import Model
    from dbgen.core.pathconstraint import Path as AP
    from dbgen.core.query import Query
    from dbgen.core.schemaclass import Schema

    Model, Query, Schema, PathAttr, AP
"""
Components of a schema: Objects, Attributes, and Relations

Also the RelTup container class is defined
"""
########################################################################################
######################
# Simple Tuple types #
######################


class AttrTup(Base):
    def __init__(self, name: str, obj: str) -> None:
        self.name = name
        self.obj = obj
        super().__init__()

    def __str__(self) -> str:
        return self.obj + "." + self.name

    def __call__(self, x: "AP" = None) -> "PathAttr":
        from dbgen.core.expr.pathattr import PathAttr

        return PathAttr(x, self)

    @classmethod
    def _strat(cls) -> "SearchStrategy":
        from hypothesis.strategies import builds

        return builds(cls)


class RelTup(Base):
    """
    A tuple (objectname, relationname) that an object can produce independent
    of a Model instance .... it will later have to be validated by the model to
    be a relation that actually exists, but lower level classes can work with it
    in the meantime
    """

    def __init__(self, obj: str, rel: str) -> None:
        self.obj = obj.lower()
        self.rel = rel.lower()
        super().__init__()

    def __str__(self) -> str:
        return f"Rel({self.obj},{self.rel})"

    @classmethod
    def _strat(cls) -> "SearchStrategy":
        from hypothesis.strategies import builds

        return builds(cls)


################################################################################
class Attr(Base):
    """ Attribute, considered from a schema-making perspective (NOT as an Expr) """

    def __init__(
        self,
        name: str,
        dtype: SQLType = None,
        identifying: bool = False,
        default: Any = None,
        desc: str = "<No description>",
        index: bool = False,
        partition_on: bool = False,
        partition_values: Sequence = None,
        partition_default: bool = True,
    ) -> None:
        assert name
        self.name = name.lower()
        self.desc = desc
        self.dtype = dtype or Int()
        self.identifying = identifying
        self.index = index
        self.default = default
        # Partition Columns
        # Use this Attribute for partitioning the Entity
        self.partition_on = partition_on
        if partition_on:
            if not isinstance(self.dtype, (Text, Int, Varchar)):
                raise NotImplementedError(
                    f"Partitioning only implemented for Int, Text, and Varchar:\n{self.name}\n{dtype}"
                )
        # Set the partitioned values
        if partition_on and not (partition_values or partition_default):
            raise DBgenMissingInfo("Cannot partition column without default partition of default values")
        # Check values can be cast to dtype
        if partition_values:
            for val in partition_values:
                try:
                    self.dtype.cast(val)
                except ValueError:
                    raise DBgenInvalidArgument(f"Invalid partition value for dtype {dtype}: \n{val}")
        self.partition_values = partition_values or []
        self.partition_default = partition_default
        super().__init__()

    def __str__(self) -> str:
        return f"Attr<{self.name},{self.dtype}>"

    ####################
    # Public methods #
    ####################

    def create_col(self, tabname: str) -> T[str, str, str]:
        """
        Create statement for column when creating a table.
        """
        dt = str(self.dtype)
        dflt = "" if self.default is None else f"DEFAULT {self.default}"
        desc = 'comment on column "{}"."{}" is \'{}\''.format(
            tabname, self.name, self.desc.replace("'", "''")
        )

        fmt_args = [self.name, dt, dflt]
        create = '"{}" \t{} {}'.format(*fmt_args)
        index = (
            f'\nCREATE INDEX IF NOT EXISTS {tabname}_{self.name}_idx ON {tabname} ("{self.name}")'
            if self.index
            else ""
        )
        return create, desc, index


class View(Base, metaclass=ABCMeta):
    def __str__(self) -> str:
        return f"View({self.name})"

    name = ""

    @abstractmethod
    def dep(self) -> Dep:
        raise NotImplementedError

    @abstractmethod
    def qstr(self) -> str:
        """String representing the query (either JSON or raw SQL)"""
        raise NotImplementedError

    def add(self, cxn: Conn) -> int:
        """add view to metadb, return PK"""
        # Try to find an Object with an equivalent hash in the existing table
        get_v = mkSelectCmd("view", ["view_id"], ["view_id"])
        v_id = sqlselect(cxn, get_v, [self.hash])

        if v_id:
            return v_id[0][0]  # already there
        else:
            # Create a new record in the View table and get its ID
            cmd = mkInsCmd("view", ["view_id", "name", "query"])
            sqlexecute(cxn, cmd, [self.hash, self.name, self.qstr()])
            return sqlselect(cxn, get_v, [self.hash])[0][0]

    @classmethod
    @abstractmethod
    def _strat(cls) -> "SearchStrategy":
        raise NotImplementedError

    def create(self) -> str:
        """
        Generate SQL necessary to create an object's corresponding table
        """
        return f"CREATE VIEW {self.name} AS {self.qstr()}"


class QView(View):
    def __init__(self, name: str, q: "Query") -> None:
        self.name = name
        self.q = q

    def qstr(self) -> str:
        return self.q.toJSON()

    def dep(self) -> Dep:
        cd = [f"{a.obj}.{a.name}" for a in self.q.allattr()]
        nc = [f"{self.name}.{x}" for x in self.q.exprs]
        return Dep(self.q.allobj(), cd, [self.name], nc)

    @classmethod
    def _strat(cls) -> "SearchStrategy":
        return builds(cls)


class RawView(View):
    def __init__(self, name: str, q: str, deps: L[str] = None, new: L[str] = None) -> None:
        self.name = name
        self.raw = q
        self.deps = deps or []
        self.new = new or []

    def qstr(self) -> str:
        return self.raw

    def dep(self) -> Dep:
        td = [d for d in self.deps if "." not in d]
        cd = [d for d in self.deps if "." in d]
        nc = [f"{self.name}.{x}" for x in self.new]
        return Dep(td, cd, [self.name], nc)

    @classmethod
    def _strat(cls) -> "SearchStrategy":
        return builds(cls)


class UserRel(Base):
    """
    USER EXPOSED Relation between objects. no need to specify source, as it is
    declared from within the UserObj constructor.
    Can be identifying or non-identifying
    """

    def __init__(
        self,
        name: str,
        tar: str = None,
        identifying: bool = False,
        desc: str = "<No description>",
        delete_trigger: str = "NO ACTION",
    ) -> None:
        self.name = name.lower()
        self.desc = desc
        self.tar = tar.lower() if tar else self.name
        self.identifying = identifying
        self.delete_trigger = delete_trigger
        super().__init__()

    def __str__(self) -> str:
        idstr = "(id)" if self.identifying else ""
        return f"{self.name}{idstr} -> {self.tar}"

    def to_rel(self, obj: str) -> "Rel":
        return Rel(
            name=self.name,
            o1=obj,
            o2=self.tar,
            identifying=self.identifying,
            desc=self.desc,
            delete_trigger=self.delete_trigger,
        )


class Entity(Base):
    """Object with attributes. Basic entity of a model"""

    def __init__(
        self,
        name: str,
        desc: str = None,
        attrs: L[Attr] = None,
        fks: L[UserRel] = None,
        id_str: str = None,
    ) -> None:

        self.name = name.lower()
        self.desc = desc or "<No description>"
        self.attrs = attrs or []
        self.fks = fks or []
        self.id_str = id_str if id_str else self.name + "_id"
        # Validate
        self.forbidden = [
            self.id_str,
            "deleted",
            "insert",
            self.name,
        ]  # RESERVED
        assert not any([a.name in self.forbidden for a in self.attrs]), (
            self.attrs,
            self.forbidden,
        )
        attr_names = [a.name for a in self.attrs]
        assert len(attr_names) == len(
            set(attr_names)
        ), f"No duplicate column names allowed, \nObject: {self.name}"
        assert all([isinstance(x, str) for x in [self.name, self.desc, self.id_str]])
        # Check at most one partitioned attr
        partitioned_attrs = list(filter(lambda x: x.partition_on, self.attrs))
        if len(partitioned_attrs) > 1:
            raise DBgenInvalidArgument("Can only have at most 1 partitioned attributes!")
        self.partition_attr = partitioned_attrs[0] if partitioned_attrs else None

        super().__init__()

    @property
    def attrdict(self) -> D[str, Attr]:
        return {a.name: a for a in self.attrs}

    @property
    def fkdict(self) -> D[str, "Rel"]:
        return {r.name: r.to_rel(self.name) for r in self.fks}

    @property
    def is_partitioned(self) -> bool:
        return self.partition_attr is not None

    def __str__(self) -> str:
        return "Entity<%s, %d attrs>" % (self.name, len(self.attrs))

    def __call__(self, **kwargs: Any) -> Load:
        """
        Construct an Load which specifies AT LEAST how to identify this
        object (via PK or data) AND POSSIBLY more non-identifying info to update

        - Attributes and relations are referred to by name with kwargs
        - A keyword equal to the object's own name signifies a PK argument
        """
        invalid_args = list(
            filter(
                lambda keyval: keyval[0] != "insert" and not isinstance(keyval[1], (ArgLike, Load)),
                kwargs.items(),
            )
        )
        if invalid_args:
            raise DBgenInvalidArgument(f"Non ArgLike kwargs provided: {invalid_args}")
        kwargs = {k.lower(): v for k, v in kwargs.items()}  # convert keys to L.C.
        pk = kwargs.pop(self.name, None)
        insert = kwargs.pop("insert", False)

        if not pk:  # if we don't have a PK reference
            err = "Cannot refer to a row in {} without a PK or essential data." " Missing essential data: {}"
            missing = set(self.ids() + self.id_fks()) - set(kwargs)
            if missing:
                raise DBgenMissingInfo(err.format(self.name, missing))

        attrs = {k: v for k, v in kwargs.items() if k in self.attrnames()}

        fks = {k: v for k, v in kwargs.items() if k not in attrs}
        for fk in fks:
            assert fk in self.fkdict, f'unknown "{fk}" kwarg in Load of {self}'
        for k, v in fks.items():
            if not isinstance(v, Load):
                # We OUGHT have a reference to a FK from a query
                assert isinstance(v, (Arg, Const))
                rel = self.fkdict[k]
                # Check for relations with names that are different from table_names
                if rel.o2 != k:
                    fks[k] = Load(obj=rel.o2, attrs={}, fks={}, pk=v)
                else:
                    fks[k] = Load(obj=k, attrs={}, fks={}, pk=v)
        # Need partition attribute name for resolving conflicts during insertion
        partition_attr = self.partition_attr.name if self.partition_attr else None
        return Load(self.name, attrs=attrs, fks=fks, pk=pk, insert=insert, partition_attr=partition_attr)

    def __getitem__(self, key: str) -> AttrTup:
        if key in self.attrdict:
            return AttrTup(key, self.name)
        else:
            raise KeyError(key + f" not found in {self}")

    # Public methods #

    def get(self, key: str) -> AttrTup:
        """
        A version of __getitem__ that doesn't check whether attribute is defined
        Use when we need to refer to an attribute which may not (yet) exist
        """
        return AttrTup(key, self.name)

    def act(self, **kwargs: Any) -> Load:
        """Do we need to add a "insert" flag in order to say: "it's ok for this
        load to insert any required parent objects recursively?"""
        return self(**kwargs)

    def r(self, relname: str) -> RelTup:
        """
        Refer to a relation of an object. Without a Model, we have to do with
        reference by name
        """
        return RelTup(self.name, relname)

    def attrnames(self, init: bool = False) -> L[str]:
        """Names of all (top-level) attributes"""
        return [a.name for a in self.attrs if a.identifying or not init]

    def create(self) -> L[str]:
        """
        Generate SQL necessary to create an object's corresponding table
        """
        create_str = f'CREATE TABLE IF NOT EXISTS "{self.name}" '
        if len(self.attrs) != 0:
            cols, coldescs, colindexes = map(
                lambda x: list(x),
                zip(*[a.create_col(self.name) for a in self.attrs]),
            )
        else:
            cols, coldescs, colindexes = [], [], []
        pk = self.id_str + " BIGINT "

        deld = "deleted BOOLEAN NOT NULL DEFAULT FALSE"
        full_cols = [pk, deld] + list(cols)
        partition_name = self.partition_attr.name if self.partition_attr else ""
        primary_key = f",PRIMARY KEY ({self.id_str}"
        primary_key += f",\"{partition_name}\")" if partition_name else ")"

        columns = "\n\t,".join(full_cols)
        partition_str = f" PARTITION BY LIST (\"{self.partition_attr.name}\")" if self.partition_attr else ""
        create_cmd = f"{create_str}\n\t({columns}\n\t{primary_key}\n){partition_str}"
        table_description = "comment on table \"{}\" is '{}'".format(self.name, self.desc.replace("'", "''"))
        sql_statements = (
            [create_cmd, table_description]
            + self.create_partitions()
            + list(coldescs)
            + list(filter(lambda x: x != "", colindexes))
        )
        return sql_statements

    def create_partitions(self) -> L[str]:
        if not self.partition_attr:
            return []
        partition_statements = []
        part_name = self.partition_attr.name
        # Add default partition if flag set
        if self.partition_attr.partition_default:
            partition_statements.append(
                f"CREATE TABLE IF NOT EXISTS \"{self.name}__{part_name}__default\" partition of \"{self.name}\" default"
            )
        # For each val, attach a  partition
        if self.partition_attr.partition_values:
            for val in self.partition_attr.partition_values:
                partition_statements.append(
                    f"CREATE TABLE IF NOT EXISTS \"{self.name}__{part_name}__{val}\" partition of \"{self.name}\" for values in ({val})"
                )
        return partition_statements

    def get_partition(self, partition_value: Any = None, default: bool = False) -> 'Partition':
        if self.partition_attr is None:
            raise ValueError(f"Entity {self.name} is not partitioned!")
        elif default:
            Partition(
                None,
                default=True,
                name=self.name,
                desc=self.desc,
                attrs=self.attrs,
                fks=self.fks,
                id_str=self.id_str,
            )
        elif partition_value not in self.partition_attr.partition_values and not default:
            raise ValueError(
                f"Requested partition value is not declared on the Partition Attribute:"
                f"\nAttr: {self.partition_attr.name}"
                f"\nPartition Values: {self.partition_attr.partition_values}"
                f"\nRequested Value: {partition_value}"
            )

        return Partition(
            partition_value,
            default=False,
            name=self.name,
            desc=self.desc,
            attrs=self.attrs,
            fks=self.fks,
            id_str=self.id_str,
        )

    def get_all_partitions(self) -> L['Partition']:
        if not self.partition_attr:
            return []
        partitions = []
        kwargs = dict(
            name=self.name,
            desc=self.desc,
            attrs=self.attrs,
            fks=self.fks,
            id_str=self.id_str,
        )
        if self.partition_attr.partition_default:
            partitions.append(Partition(None, default=True, **kwargs))  # type: ignore
        partitions.extend(map(lambda val: self.get_partition(val), self.partition_attr.partition_values))

        return partitions

    def id(self, path: "AP" = None) -> PK:
        """Main use case: GROUP BY an object, rather than a particular column"""
        from dbgen.core.expr.pathattr import PathAttr

        return PK(PathAttr(path, AttrTup(self.id_str, self.name)))

    def ids(self) -> L[str]:
        """Names of all the identifying (top-level) attributes."""
        return [a.name for a in self.attrs if a.identifying]

    def id_fks(self) -> L[str]:
        """Names of all the identifying (top-level) FKs """
        return [f.name for f in self.fks if f.identifying]

    def add(self, cxn: Conn) -> int:
        """
        Add this Object to a metaDB that stores information about a model (if
        it's not already there), and return the ID.
        """
        # Try to find an Object with an equivalent hash in the existing table
        get_t = mkSelectCmd("object", ["object_id"], ["object_id"])
        t_id = sqlselect(cxn, get_t, [self.hash])

        if t_id:
            return t_id[0][0]  # already there
        else:
            # Create a new record in the Object table and get its ID
            name = self.name
            cmd = mkInsCmd("object", ["object_id", "name", "description"])
            sqlexecute(cxn, cmd, [self.hash, name, self.desc])
            tab_id = sqlselect(cxn, get_t, [self.hash])[0][0]

            # Before returning ID, we have to populate Attr table
            ins_cols = [
                "object",
                "name",
                "dtype",
                "description",
                "defaultval",
                "attr_id",
            ]

            for c in self.attrs:
                # Insert info about an attribute
                binds = [
                    tab_id,
                    c.name,
                    str(c.dtype),
                    c.desc,
                    str(c.default),
                    c.hash,
                ]

                cmd = mkInsCmd("attr", ins_cols)
                sqlexecute(cxn, cmd, binds)

            return tab_id

    def rename_attr(self, aname: str, newname: str) -> "Entity":
        """ Copy of object with a renamed attribute """
        o = self.copy()
        newattr = o.attrdict[aname]
        o.attrs.remove(newattr)
        newattr.name = newname
        o.attrs.append(newattr)
        return o


class Partition(Entity):
    """Sub-Entity for a specific List Partition"""

    def __init__(
        self,
        partition_value: Any,
        default: bool,
        name: str,
        desc: str,
        attrs: L[Attr],
        fks: L[UserRel],
        id_str: str,
    ) -> None:
        self.partition_value = partition_value
        self.default = default
        super().__init__(name, desc=desc, attrs=attrs, fks=fks, id_str=id_str)
        assert self.partition_attr, "Trying to make partition out of unpartitioned Entity!"
        # Set name of Partition Entity
        if default:
            self.name = f"{self.name}__{self.partition_attr.name}__default"
            self._parent_name = name
        else:
            assert partition_value in self.partition_attr.partition_values
            self.name = f"{self.name}__{self.partition_attr.name}__{self.partition_value}"
            self._parent_name = name


class Rel(Base):
    """
    Asymmetric Relation between objects

    Can be identifying or non-identifying
    """

    _valid_triggers = ("NO ACTION", "CASCADE", "SET NULL")

    def __init__(
        self,
        name: str,
        o1: str,
        o2: str = None,
        identifying: bool = False,
        desc: str = "<No description>",
        delete_trigger: str = "NO ACTION",
    ) -> None:
        self.name = name.lower()
        self.desc = desc
        self.o1 = o1.lower()
        self.o2 = o2.lower() if o2 else self.name
        self.identifying = identifying
        if delete_trigger:
            assert (
                delete_trigger in self._valid_triggers
            ), f"Invalid delete trigger value {delete_trigger} please use one of: {self._valid_triggers}"
            self.delete_trigger = delete_trigger
        super().__init__()

    def __str__(self) -> str:
        return "Rel<{}{},{} -> {}>".format(
            self.name,
            "(id)" if self.identifying else "",
            self.o1,
            self.o2,
        )

    def __repr__(self) -> str:
        return self.__str__()

    # Public methods #

    def print(self) -> str:
        return f"{self.o1}.{self.name}"

    def tup(self) -> RelTup:
        """Throw away info to make a RelTup"""
        return RelTup(self.o1, self.name)

    @property
    def default(self) -> bool:
        """Whether or not this Relation is the 'default' one"""
        return self.name == self.o2

    @property
    def objs(self) -> S[str]:
        return {self.o1, self.o2}

    def other(self, obj: str) -> str:
        """
        Often we don't know which direction we are traversing on a FK
        We know which end we're coming from and simply want the other end
        """
        assert obj in self.objs, f"{obj} not found in {self}"

        out = [o for o in self.objs if o != obj]

        if out:
            return out[0]  # normal case
        else:
            return obj  # both to and from are the same table!

    # Private Methods #
    # ...


class SuperRel(Base):
    """
    Contains information about source table name, relation name,
    and target table ID column name.

    Example:
    SuperRel: Person.favorite_book <-> book.title
        name: favorite_book
        source: Person
        book: target
        target_id_str: Title
    """

    def __init__(
        self,
        name: str,
        source: str,
        target: str,
        target_id_str: str,
        identifying: bool = False,
        desc: str = "<No description>",
    ) -> None:
        self.name = name.lower()
        self.source = source.lower()
        self.target = target.lower()
        self.target_id_str = target_id_str
        self.identifying = identifying
        self.desc = desc

        super().__init__()

    def __str__(self) -> str:
        return "SuperRel<{},{} -> {}.{}>".format(
            self.name,
            self.source,
            self.target,
            self.target_id_str,
        )

    def __repr__(self) -> str:
        return self.__str__()

    # Public methods #

    def print(self) -> str:
        return f"{self.source}.{self.name}"

    def tup(self) -> RelTup:
        """Throw away info to make a RelTup"""
        return RelTup(self.source, self.name)

    @property
    def default(self) -> bool:
        """Whether or not this Relation is the 'default' one"""
        return self.name == self.target

    @property
    def objs(self) -> S[str]:
        return {self.source, self.target}

    def other(self, obj: str) -> str:
        """
        Often we don't know which direction we are traversing on a FK
        We know which end we're coming from and simply want the other end
        """
        assert obj in self.objs, f"{obj} not found in {self}"

        out = [o for o in self.objs if o != obj]

        if out:
            return out[0]  # normal case
        else:
            return obj  # both to and from are the same table!

    def to_rel(self) -> Rel:
        return Rel(self.name, self.source, self.target, self.identifying, self.desc)

    # Private Methods #
    # ...
    @classmethod
    def _strat(cls) -> "SearchStrategy":
        return builds(cls)

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            raise ValueError(f"Comparing {type(other)} to type {self.__class__}")
        return repr(self) < repr(other)


class Path(Base):
    """
    Some list of foreign keys (possibly empty), possibly followed by attribute
    """

    def __init__(self, rels: L[RelTup] = None, attr: AttrTup = None) -> None:
        self.rels = rels or []
        self.attr = attr
        assert rels or attr
        super().__init__()

    def __str__(self) -> str:
        p = f"[{','.join(map(str, self.rels))}]" if self.rels else ""
        comma = "," if self.rels else ""
        a = comma + str(self.attr) if self.attr else ""
        return f"Path({p}{a})"

    @classmethod
    def _strat(cls) -> "SearchStrategy":
        return builds(cls, rels=lists(RelTup._strat(), max_size=2), attr=AttrTup._strat())

    def start(self) -> str:
        """Starting point of a path, always an object (name is returned)"""
        if self.rels:
            return self.rels[0].obj
        else:
            assert self.attr
            return self.attr.obj

    def select(self, m: "Model") -> str:
        """ Thing to select for when comparing path equality """
        alias = self.start()
        for r in self.rels:
            alias += "$" + r.rel
        col = self._end_attr(m).name
        return f'"{alias}"."{col}"'

    def joins(self, ids: D[str, str], m: "Model") -> L[str]:
        """From clause that makes self.select() defined in query"""

        j = []  # type: L[str]
        oldpath = self.start()
        jstr = 'JOIN "{0}" AS "{1}" ON "{1}"."{4}" = "{2}"."{3}"'
        next = self.start()
        for i, r in enumerate(self.rels[:-1]):
            newpath = oldpath + "$" + r.rel
            next = self.rels[i + 1].obj
            nextid = ids[next]
            args = [next, newpath, oldpath, r.rel, nextid]
            j.append(jstr.format(*args))
            oldpath = newpath
        if self.rels:
            last = self.rels[-1].rel
            lastid = ids[m.get_rel(self.rels[-1]).o2]
            newpath = oldpath + "$" + last
            lastargs = [self._end_attr(m).obj, newpath, oldpath, last, lastid]
            j.append(jstr.format(*lastargs))
        return j

    def _end_attr(self, m: "Model") -> AttrTup:
        """the ID colname if we don't have a normal attribute as terminus"""
        if not self.attr:
            rel = m.get_rel(self.rels[-1])
            return AttrTup(m[rel.o2].id_str, rel.o2)
        else:
            return self.attr

    def _path_end(self, m: "Schema") -> str:
        """Determine the datatype of the end of a path."""
        if (not self.attr) or (AttrTup(m[self.attr.obj].id_str, self.attr.obj) == self.attr):
            return "id"
        else:
            o = m[self.attr.obj]
            a = [a for n, a in o.attrdict.items() if n == self.attr.name]
            assert len(a) < 2
            if not a:
                err = "Could not find %s in %s: Path %s"
                raise ValueError(err % (a, o, self))
            return str(a[0].dtype)


class PathEQ(Base):
    """Specification that two paths should result in the same value."""

    def __init__(self, p1: Path, p2: Path) -> None:
        assert p1 != p2, "Cannot do pathEQ between things that are literally equivalent"
        assert p1.start() == p2.start(), "Paths must have same start point"
        self.paths = {p1, p2}
        super().__init__()

    def __str__(self) -> str:
        return "PathEQ({},{})".format(*self.paths)

    def __iter__(self) -> Iter[Path]:
        return iter(self.paths)

    def __contains__(self, a: AttrTup) -> bool:
        if isinstance(a, AttrTup):
            return any([a == p.attr for p in self.paths])
        else:
            raise TypeError("add to this to support more types of searching")

    def any(self) -> Path:
        """Gives one of the paths, doesn't matter which"""
        return next(iter(self))

    def start(self) -> str:
        return self.any().start()



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import configparser

# External imports
from os import environ
from os.path import dirname, exists, join


def get_config():
    config = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
    def_pth = join(dirname(__file__), "model.cfg")
    config_pth = environ.get("MODEL_CONFIG", def_pth)
    if exists(config_pth):
        config.read(config_pth)
        return config
    else:
        raise FileNotFoundError("No Config path found")



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External Modules
from abc import ABCMeta, abstractmethod
from functools import reduce
from operator import add
from typing import TYPE_CHECKING, Any
from typing import Callable as C
from typing import List as L
from typing import Tuple as T
from typing import Union as U

from hypothesis.strategies import SearchStrategy, builds, just, lists, one_of

# Internal Modules
from dbgen.core.expr.sqltypes import Boolean, Decimal, Int, SQLType, Text, Varchar
from dbgen.utils.lists import concat_map
from dbgen.utils.misc import Base, anystrat

if TYPE_CHECKING:
    from dbgen.core.expr.pathattr import PathAttr

    PathAttr


"""
Python-sql interface.
"""

###############################################################################
Fn = C[[Any], str]  # type shortcut


class Expr(Base, metaclass=ABCMeta):

    # Constants
    # ----------
    @property
    def agg(self) -> bool:
        return False  # by default, we assume not an Aggregation

    # Abstract methods
    # -----------------
    @abstractmethod
    def fields(self) -> list:
        """
        List of immediate substructures of the expression (not recursive)
        """
        raise NotImplementedError

    @abstractmethod
    def show(self, f: Fn) -> str:
        """
        Apply function recursively to fields
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _strat(cls) -> SearchStrategy:
        abc = Literal("abc")
        exprs: L[Expr] = [
            One,
            LT(One, One),
            GT(One, Zero),
            LEN(abc),
            EQ(One, One),
            LEFT(abc, One),
            RIGHT(abc, One),
        ]
        return one_of(*[just(x) for x in exprs])

    # --------------------------#
    # Representing expressions #
    # --------------------------#
    def __str__(self) -> str:
        """
        Default string representation: all fields run through str()
        """
        return self.show(lambda x: str(x))

    def __repr__(self) -> str:
        return f"Expr<{str(self)}>"

    def __hash__(self) -> int:
        return hash(str(self))

    # --------------------#
    # Overloaded methods #
    # --------------------#
    def __abs__(self) -> "ABS":
        return ABS(self)

    def __add__(self, other: "Expr") -> "PLUS":
        return PLUS(self, other)

    def __mul__(self, other: "Expr") -> "MUL":
        return MUL(self, other)

    def __pow__(self, other: "Expr") -> "POW":
        return POW(self, other)

    def __sub__(self, other: "Expr") -> "MINUS":
        return MINUS(self, other)

    def __or__(self, other: Any) -> "Expr":
        raise NotImplementedError

    del __or__  # tricking the type checker to use |Infix|

    def __truediv__(self, other: "Expr") -> "DIV":
        return DIV(self, other)

    # -----------------#
    # Private Methods #
    # -----------------#

    def _all(self) -> list:
        """
        List of all fundamental things that are involved in the expression
        Recursively expand substructures and flatten result
        """
        return concat_map(self.get_all, self.fields())

    # Static methods #

    @staticmethod
    def get_all(x: Any) -> list:
        """If it has any recursive structure to unpack, unpack it"""
        return x._all() if hasattr(x, "_all") else [x]


################################################################################

##############
# Subclasses #
##############


class Unary(Expr):
    """
    Expression that depends on just one individual thing
    """

    # Input-indepedent parameters
    # ----------------------------
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    # Implement Expr abstract methods
    # --------------------------------
    def fields(self) -> list:
        return [self.x]

    def show(self, f: Fn) -> str:
        x = f(self.x)
        return f"{self.name}({x})"

    # Class-specific init
    # -------------------
    def __init__(self, x: Expr) -> None:
        assert isinstance(x, Expr), (x, type(x))
        self.x = x

    @classmethod
    def _strat(cls, strat: SearchStrategy = None) -> SearchStrategy:
        from dbgen.core.expr.exprstrat import exprstrat

        x = exprstrat if strat is None else strat
        assert isinstance(x, SearchStrategy)
        return builds(cls, x=x)


class Binary(Expr):
    """
    Expression that depends on two individual things
    """

    # Input-indepedent parameters
    # ----------------------------
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    infix = True

    # Implement Expr abstract methods
    # --------------------------------
    def fields(self) -> list:
        return [self.x, self.y]

    def show(self, f: Fn) -> str:
        x, y = f(self.x), f(self.y)
        if self.infix:
            return f"({x} {self.name} {y})"
        else:
            return f"{self.name}({x},{y})"

    # Class-specific init
    # -------------------
    def __init__(self, x: Expr, y: Expr) -> None:
        assert all([isinstance(a, Expr) for a in [x, y]]), [
            x,
            type(x),
            y,
            type(y),
        ]
        self.x, self.y = x, y

    @classmethod
    def _strat(cls, strat: SearchStrategy = None) -> SearchStrategy:
        from dbgen.core.expr.exprstrat import exprstrat

        x = exprstrat() if strat is None else strat
        return builds(cls, x=x, y=x)


class Ternary(Expr):
    """
    Expression that depends on three individual things
    """

    # Input-indepedent parameters
    # ----------------------------
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    # Implement Expr abstract methods
    # --------------------------------
    def fields(self) -> list:
        return [self.x, self.y, self.z]

    def show(self, f: Fn) -> str:
        x, y, z = f(self.x), f(self.y), f(self.z)
        return f"{self.name}({x},{y},{z})"

    # Class-specific init
    # -------------------
    def __init__(self, x: Expr, y: Expr, z: Expr) -> None:
        assert all([isinstance(a, Expr) for a in [x, y, z]])
        self.x, self.y, self.z = x, y, z

    @classmethod
    def _strat(cls, strat: SearchStrategy = None) -> SearchStrategy:
        from dbgen.core.expr.exprstrat import exprstrat

        x = exprstrat() if strat is None else strat
        return builds(cls, x=x, y=x, z=x)


class Nary(Expr):
    """
    SQL Functions that take multiple arguments, initialized by user with
    multiple inputs (i.e. not a single list input)
    """

    # Input-indepedent parameters
    # ----------------------------
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    def delim(self) -> str:
        return ","  # default delimiter

    def fields(self) -> list:
        return self.xs

    def __init__(self, xs: L[Expr]) -> None:
        self.xs = xs
        assert all([isinstance(x, Expr) for x in xs])

    def show(self, f: Fn) -> str:
        xs = map(f, self.xs)
        d = f" {self.delim} "
        return f"{self.name}({d.join(xs)})"

    @classmethod
    def _strat(cls, strat: SearchStrategy = None) -> SearchStrategy:
        from dbgen.core.expr.exprstrat import exprstrat

        x = exprstrat() if strat is None else strat
        return builds(cls, xs=lists(x, min_size=1, max_size=2))


class Named(Expr):
    """
    Inherit from this to allow any arbitrary class, e.g. XYZ(object),
     to automatically deduce that its 'name' property should be 'XYZ'
    """

    @property
    def name(self) -> str:
        return type(self).__name__


class Agg(Expr):
    """
    This class is meant to be inherited by any SQL function we want to flag
    as an aggregation.

    We can optionally specify what objects we want to aggregate over, otherwise
    the intent will be guessed
    """

    @property
    def agg(self) -> bool:
        return True

    def __init__(self, x: Expr, objs: list = None) -> None:
        assert issubclass(type(x), Expr)
        self.x = x
        self.objs = objs or []


################################################################################
# Specific Expr classes for user interface
###########################################

# Ones that work out of the box
# ------------------------------
class ABS(Named, Unary):
    pass


class SQRT(Named, Unary):
    pass


class MAX(Named, Agg, Unary):
    pass


class SUM(Named, Agg, Unary):
    pass


class MIN(Named, Agg, Unary):
    pass


class AVG(Named, Agg, Unary):
    pass


class COUNT(Agg, Named, Unary):
    pass


class CONCAT(Named, Nary):
    pass


class BINARY(Named, Unary):
    pass  # haha


class REGEXP(Named, Binary):
    pass


class REPLACE(Named, Ternary):
    pass


class COALESCE(Named, Nary):
    pass


class LIKE(Named, Binary):
    pass


# Ones that need a field defined
# -------------------------------
class Tup(Nary):
    name = ""


class LEN(Unary):
    name = "CHAR_LENGTH"


class MUL(Binary):
    name = "*"


class DIV(Binary):
    name = "/"


class PLUS(Binary):
    name = "+"


class MINUS(Binary):
    name = "-"


class POW(Named, Binary):
    infix = False


class LEFT(Named, Binary):
    infix = False


class RIGHT(Named, Binary):
    infix = False


class JSON_EXTRACT(Named, Binary):
    infix = False


class EQ(Binary):
    name = "="


class NE(Binary):
    name = "!="


class LT(Binary):
    name = "<"


class GT(Binary):
    name = ">"


class LE(Binary):
    name = "<="


class GE(Binary):
    name = ">="


class OR(Nary):
    """ Can be used as a binary operator (|OR|) or as a function OR(a,b,...)"""

    name = ""
    delim = "OR"


class AND(Nary):
    name = ""
    delim = "\n\tAND"


class And(Nary):
    name = ""
    delim = "\n\tAND"


class NOT(Named, Unary):
    wrap = False


class NULL(Named, Unary):
    def show(self, f: Fn) -> str:
        return f"{f(self.x)} is NULL"


class ARRAY(Named, Nary):
    def show(self, f: Fn) -> str:
        xs = map(f, self.xs)
        d = f" {self.delim} "
        return f"{self.name}[{d.join(xs)}]"


# Ones that need to be implemented from scratch
# ----------------------------------------------


class Literal(Expr):
    def __init__(self, x: Any) -> None:
        self.x = x

    def fields(self) -> L[Expr]:
        return []

    def show(self, f: Fn) -> str:

        if isinstance(self.x, str):
            return "('%s')" % f(self.x).replace("'", "\\'").replace("%", "%%")
        elif self.x is None:
            return "(NULL)"
        else:
            x = f(self.x)
            return f"({x})"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls, x=anystrat)


class IN(Named):
    def __init__(self, x: Expr, xs: U[L[Expr], L[Literal]]) -> None:
        self.x = x
        self.xs = xs

    def fields(self) -> L[Expr]:
        return [self.x] + self.xs  # type: ignore

    def show(self, f: Fn) -> str:
        xs = map(f, self.xs)
        return f"{f(self.x)} IN ({','.join(xs)})"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


##########
class CASE(Expr):
    def __init__(self, cases: L[T[Expr, Expr]], else_: Expr) -> None:
        self.cases = cases
        self.else_ = else_

    def fields(self) -> L[Expr]:
        k, v = map(lambda x: list(x), zip(*self.cases))
        return k + v + [self.else_]

    def show(self, f: Fn) -> str:
        body = " ".join(["WHEN ({}) THEN ({})".format(f(k), f(v)) for k, v in self.cases])
        end = f" ELSE ({f(self.else_)}) END"
        return "CASE  " + body + end

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class IF_ELSE(Expr):
    def __init__(self, cond: Expr, if_: Expr, else_: Expr) -> None:
        self.cond = cond
        self.if_ = if_
        self.else_ = else_

    def fields(self) -> L[Expr]:
        return [self.cond, self.if_, self.else_]

    def show(self, f: Fn) -> str:
        c, i, e = map(f, self.fields())
        return f"CASE WHEN ({c}) THEN ({i}) ELSE ({e}) END"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class CONVERT(Expr):
    def __init__(self, expr: Expr, dtype: SQLType) -> None:
        self.expr = expr
        self.dtype = dtype

        err = "Are you SURE that Postgres can convert to this dtype? %s\n Also did you make sure to use an instance and not the class itself?"
        assert isinstance(dtype, (Decimal, Varchar, Text, Int, Boolean)), err % dtype

    def fields(self) -> L[Expr]:
        return [self.expr]

    def show(self, f: Fn) -> str:
        e = f(self.expr)
        return f"CAST({e} AS {self.dtype})"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class SUBSELECT(Expr):
    """Hacky way of getting in subselect ... will not automatically detect
    dependencies."""

    def __init__(self, expr: Expr, tab: str, where: str = "1") -> None:
        self.expr = expr
        self.tab = tab
        self.where = where

    def fields(self) -> L[Expr]:
        return [self.expr]

    def show(self, f: Fn) -> str:
        e = f(self.expr)
        return f"(SELECT {e} FROM {self.tab} WHERE {self.where} )"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class GROUP_CONCAT(Agg):
    def __init__(self, expr: Expr, delim: str = None, order: Expr = None) -> None:
        self.expr = expr
        self.delim = delim or ","
        self.order = order

    def fields(self) -> L[Expr]:
        return [self.expr] + ([self.order] if self.order is not None else [])

    @property
    def name(self) -> str:
        return "string_agg"

    def show(self, f: Fn) -> str:
        ord = "ORDER BY " + f(self.order) if self.order is not None else ""

        return f"string_agg({f(self.expr)} :: TEXT,'{self.delim}' {ord})"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class ARRAY_AGG(Agg):
    def __init__(self, expr: Expr, order: Expr = None) -> None:
        self.expr = expr
        self.order = order

    def fields(self) -> L[Expr]:
        return [self.expr] + ([self.order] if self.order is not None else [])

    @property
    def name(self) -> str:
        return "array_agg"

    def show(self, f: Fn) -> str:
        ord = "ORDER BY " + f(self.order) if self.order is not None else ""

        return f"array_agg({f(self.expr)} {ord})"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class STD(Agg, Unary):
    name = "stddev_pop"


class POSITION(Named, Binary):
    def show(self, f: Fn) -> str:
        return f"POSITION({f(self.x)} in {f(self.y)})"


##############################################################################
##############################################################################
class PK(Expr):
    """Special Expr type for providing PK + UID info"""

    def __init__(self, pk: "PathAttr") -> None:
        self.pk = pk

    @property
    def name(self) -> str:
        return "PK"

    def show(self, f: Fn) -> str:
        return f(self.pk)

    def fields(self) -> list:
        return [self.pk]

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


############################
# Specific Exprs and Funcs #
############################

Zero = Literal(0)
One = Literal(1)
true = Literal("true")
false = Literal("false")


def Sum(iterable: L[Expr]) -> Expr:
    """The builtin 'sum' function doesn't play with non-integers"""
    return reduce(add, iterable, Zero)


def R2(e1: Expr, e2: Expr) -> Expr:
    """
    Pearson correlation coefficient for two independent vars
    "Goodness of fit" for a model y=x, valued between 0 and 1
    """
    return (AVG(e1 * e2) - (AVG(e1) * AVG(e2))) / (STD(e1) * STD(e2))


def toDecimal(e: Expr) -> Expr:
    return CONVERT(e, Decimal())



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from os.path import dirname, join

from dbgen import Env, Import

ROOT = join(dirname(__file__), 'data')

defaultEnv = Env(
    [
        Import("json", ["load"]),
        Import("typing", aliased_imports={"List": "L", "Tuple": "T"}),
        Import("dbgen.example.constants", "ROOT"),
        Import("os.path", "join"),
    ]
)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from dbgen import (
    CONVERT,
    EQ,
    GT,
    IF_ELSE,
    LEN,
    MAX,
    Boolean,
    Expr,
    Generator,
    Literal,
    Model,
    One,
    PyBlock,
    Query,
    Zero,
    false,
    true,
)

################################################################################


def analysis(m: Model) -> None:
    ##################
    # Extract tables #
    ##################
    tabs = [
        "scientist",
        "procedures",
        "history",
        "electrode",
        "anode",
        "fuel_cell",
        "element",
        "electrode_composition",
    ]

    (
        Scientist,
        Procedures,
        History,
        Electrode,
        Anode,
        Fuel_cell,
        Element,
        Electrode_compostion,
    ) = map(m.get, tabs)

    rels = [
        History.r("operator"),
        Fuel_cell.r("cathode"),
        Fuel_cell.r("anode"),
        Anode.r("electrode"),
        Electrode.r("sample"),
        History.r("sample"),
        History.r("expt_type"),
    ]
    (
        history__operator,
        fuel_cell__cathode,
        fuel_cell__anode,
        anode__electrode,
        electrode__sample,
        history__sample,
        history__expt_type,
    ) = map(m.get_rel, rels)

    ################
    # ADD TO MODEL #
    ################
    ############################################################################
    ssn_div_desc = (
        "Record whether the sample processing step ID# evenly "
        "divides the ssn of the scientist who processed it, "
        "and only computes this when the procedure name is "
        "greater than three letter long!"
    )

    spath = m.make_path("scientist", [history__operator])
    ppath = m.make_path("procedures", [history__expt_type])

    ssnquery = Query(
        exprs=dict(ssn=Scientist["ssn"](spath), step=History["step"](), h=History.id()),
        basis=["history"],
        constr=GT(LEN(Procedures["procedure_name"](ppath)), Literal(3)),
    )
    ssnpb = PyBlock(
        lambda ssn, step: ssn % step == 0,
        args=[ssnquery["ssn"], ssnquery["step"]],
        outnames=["answer"],
    )

    ssn_divided = Generator(
        name="ssn_divided",
        desc=ssn_div_desc,
        loads=[History(step_divides_ssn=ssnpb["answer"], history=ssnquery["h"])],
        query=ssnquery,
        transforms=[ssnpb],
    )

    ############################################################################

    # Represent whether a battery had a calcinated anode with a SQL expression
    proc_name = Procedures["procedure_name"]

    pp = m.make_path(
        "procedures",
        [
            history__expt_type,
            history__sample,
            electrode__sample,
            anode__electrode,
            fuel_cell__anode,
        ],
    )

    calcined_anode = EQ(proc_name(pp), Literal("Calcination"))

    def bool_to_tinyint(x: Expr) -> Expr:
        """Useful utility function to generate SQL Expression"""
        return IF_ELSE(x, One, Zero)

    def int2bool(x: Expr) -> Expr:
        return IF_ELSE(EQ(x, One), true, false)

    c_query = Query(
        exprs=dict(
            f=MAX(Fuel_cell.id()),
            calc=CONVERT(int2bool(MAX(bool_to_tinyint(calcined_anode))), Boolean()),
        ),
        basis=["fuel_cell"],
        aggcols=[Fuel_cell["expt_id"]()],
    )  # Aggregate over this object

    fc_act = Fuel_cell(calc_anode=c_query["calc"], fuel_cell=c_query["f"])
    calcined = Generator(
        name="calcined",
        desc="Record whether battery anode was ever calcinated",
        query=c_query,
        loads=[fc_act],
    )

    ############################################################################
    ############################################################################
    ############################################################################

    m.add([ssn_divided, calcined])



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Common utilities for interfacing with SQL Databases."""
from io import StringIO
from random import getrandbits
from time import sleep

# External Modules
from typing import Any, List
from typing import Tuple as T

from psycopg2 import Error, ProgrammingError  # type: ignore
from psycopg2.extras import DictCursor, execute_batch  # type: ignore

Connection = Any
# ##############################################################################
# # Interface with DB
# # ------------------

###########
# Shortcuts
###########


def mkInsCmd(tabName: str, names: List[str], pk: str = None) -> str:
    """
    Creates an insert command for a table named tabName with columns named names.

    Args:
        tabName (str): name of the table to insert into
        names (List[str]): list of names of columns to insert into
        pk (str, optional): name of the primary key of the table if it is not {tabName}_id. Defaults to {tabName}_id.

    Returns:
        str: Sql statement for the insertion with correct formmatting
    """
    dup = f" ON CONFLICT ({pk or tabName + '_id'}) DO NOTHING"
    ins_names = ",".join(['"%s"' % (n) for n in names])
    fmt_args = [tabName, ins_names, ",".join(["%s"] * len(names)), dup]
    return "INSERT INTO {} ({}) VALUES ({}) {}".format(*fmt_args)


def mkUpdateCmd(tabName: str, names: List[str], keys: List[str]) -> str:
    """
    Creates an update command for a table named tabName with columns named names.

    Args:
        tabName (str): name of the table to insert into
        names (List[str]): list of names of columns to update
        keys (List[str]): where statements to update on.

    Returns:
        str: Sql statement for the update with correct formmatting
    """
    fmt_args = [tabName, addQs(names, ","), addQs(keys, " AND ")]
    return "UPDATE {} SET {} WHERE {}".format(*fmt_args)


def mkSelectCmd(tabName: str, get: List[str], where: List[str]) -> str:
    fmt_args = [",".join(get), tabName, addQs(where, " AND ")]
    return "SELECT {} FROM {} WHERE {}".format(*fmt_args)


##############################################################################


def select_dict(conn: Connection, q: str, binds: list = []) -> List[dict]:
    # print('SELECTING with: \n'+sub(q,binds))
    with conn.cursor(cursor_factory=DictCursor) as cxn:
        cxn.execute(q, vars=binds)
        return cxn.fetchall()


def sqlselect(conn: Connection, q: str, binds: list = []) -> List[tuple]:
    # print('\n\nSQLSELECT ',q)#,binds)
    with conn.cursor() as cxn:  # type: ignore
        cxn.execute(q, vars=binds)
        return cxn.fetchall()


def sqlexecute(conn: Connection, q: str, binds: list = []) -> list:
    with conn.cursor() as cxn:  # type: ignore
        while True:
            try:
                cxn.execute(q, vars=binds)
                try:
                    out = cxn.fetchall()
                    return out
                except ProgrammingError as e:
                    if "no results to fetch" in e.args:
                        return []
                    else:
                        raise Error(e)
            except Error as e:
                if e.args[0] in [1205, 1213]:  # deadlock error codes
                    print("SLEEPING")
                    sleep(10)
                elif e.pgcode in ["42701"]:
                    raise e
                else:
                    raise Error(e)


def sqlexecutemany(conn: Connection, q: str, binds: List[list] = []) -> None:
    with conn.cursor() as cxn:  # type: ignore
        while True:
            try:
                execute_batch(cur=cxn, sql=q, argslist=binds)
                # return cxn.fetchall()
                break
            except Error as e:
                if e.args[0] in [1205, 1213]:  # deadlock error codes
                    print("SLEEPING")
                    sleep(10)
                else:
                    raise Error(e)


def addQs(xs: list, delim: str) -> str:
    """
    Ex: ['a','b','c'] + ',' ==> 'a = %s, b = %s, c = %s'
    """
    return delim.join([f"{x} = %s" for x in xs])


def batched_cursor(cursor: Any, arraysize: int = 1000) -> Any:
    """An iterator that uses fetchmany to keep memory usage down"""
    while True:
        results = cursor.fetchmany(arraysize)
        if not results:
            break
        yield from results


def fast_load(
    conn: Connection,
    rows: T[List[Any], ...],
    table_name: str,
    col_names: List[str],
    obj_pk_name: str,
) -> None:

    # write rows to string io object to allow copy_from to be used
    io_obj = StringIO()
    for row in rows:
        io_obj.write("\t".join(map(str, row)) + "\n")

    io_obj.seek(0)
    # Temporary table to copy data into
    # Set name to be hash of input rows to ensure uniqueness for parallelization
    temp_table_name = table_name + "_temp_load_table_" + str(getrandbits(64))

    # Need to create a temp table to copy data into
    # Add an auto_inc column so that data can be ordered by its insert location
    create_temp_table = """
        DROP TABLE IF EXISTS {temp_table_name};
        CREATE TEMPORARY TABLE {temp_table_name} AS
        TABLE {obj}
        WITH NO DATA;
        ALTER TABLE {temp_table_name}
        ADD COLUMN auto_inc SERIAL NOT NULL;
        """.format(
        obj=table_name, temp_table_name=temp_table_name
    )
    from dbgen.templates import jinja_env

    insert_template = jinja_env.get_template("insert.sql.jinja")
    template_args = dict(
        obj=table_name,
        obj_pk_name=obj_pk_name,
        temp_table_name=temp_table_name,
        all_column_names=col_names,
        first=False,
        update=True,
    )
    insert_statement = insert_template.render(**template_args)
    with conn.cursor() as curs:
        curs.execute(create_temp_table)
        curs.copy_from(io_obj, temp_table_name, null="None", columns=col_names)
        curs.execute(insert_statement)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Configure the dbgen logger for each run"""
# External imports
import logging
import logging.config
import sys
from logging import Logger
from logging.handlers import RotatingFileHandler
from pathlib import Path


def setup_logger(
    logger_name: str,
    level: int,
    write_logs: bool,
    log_path: Path,
    log_format: str,
    log_to_stdout: bool = False,
) -> Logger:
    """
    configures the dbgen logger for a given model

    Args:
        logger_name (str, optional): name of the logger. Defaults to "".
        level (int, optional): level of the logger. Defaults to logging.INFO.
        write_logs (bool, optional): should the log be written to a file. Defaults to False.
        log_path (Path, optional): if write_logs write logs to this path.
        Defaults to default_log_path.

    Returns:
        Logger: the Logger object
    """
    format = logging.Formatter(log_format)
    custom_logger = logging.getLogger(logger_name)
    custom_logger.setLevel(level)
    custom_logger.propagate = False
    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setFormatter(format)
    # Only log to screen if flag is set
    console_level = level if log_to_stdout else logging.ERROR
    console_handler.setLevel(console_level)
    custom_logger.addHandler(console_handler)
    if write_logs:
        log_path.parent.mkdir(exist_ok=True, parents=True)
        info_handler = RotatingFileHandler(log_path, maxBytes=10485760, backupCount=1)
        info_handler.setLevel(level)
        info_handler.setFormatter(format)
        custom_logger.addHandler(info_handler)

    return custom_logger



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = ["version"]
from os.path import dirname, exists, join

version = "0.4.6"

try:
    curr_dir = dirname(__file__)
    git_ver_file = join(curr_dir, "git_version")
    if exists(git_ver_file):
        with open(git_ver_file) as f:
            git_version: str = f.read().strip()
    else:
        git_version = ""
except FileNotFoundError:
    git_version = ""



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Internal Modules
from dbgen import Attr, Decimal, Entity, Model, Rel, Varchar

from .generators import add_generators  # type: ignore

emp_rels = [Rel("department"), Rel("manager", "employee")]
emp = Entity(
    name="employee",
    desc="an example entity",
    attrs=[
        Attr("name", Varchar(), desc="Full name", identifying=True),
        Attr("salary", Decimal(), desc="USD/yr"),
    ],
    fks=emp_rels,
)


dept = Entity(
    name="department",
    desc="collection of employees",
    attrs=[
        Attr("name", Varchar(), identifying=True),
        Attr("total_salary", Decimal()),
    ],
)


################################################################################
################################################################################
################################################################################

objs = [
    dept,
    emp,
]


def make_model(model_name: str) -> Model:
    m = Model(model_name)
    m.add(objs)
    add_generators(m)
    return m
